# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Plan merge implementation for versioned files."""

import logging

import patiencediff
from vcsgraph import graph as _mod_graph

from . import revision as _mod_revision

logger = logging.getLogger(__name__)
from vcsgraph.tsort import merge_sort

from . import weave
from .errors import RevisionNotPresent


class _PlanMergeBase:
    def __init__(self, a_rev, b_rev, vf, key_prefix):
        """Contructor.

        :param a_rev: Revision-id of one revision to merge
        :param b_rev: Revision-id of the other revision to merge
        :param vf: A VersionedFiles containing both revisions
        :param key_prefix: A prefix for accessing keys in vf, typically
            (file_id,).
        """
        self.a_rev = a_rev
        self.b_rev = b_rev
        self.vf = vf
        self._last_lines = None
        self._last_lines_revision_id = None
        self._cached_matching_blocks = {}
        self._key_prefix = key_prefix
        self._precache_tip_lines()

    def _precache_tip_lines(self):
        lines = self.get_lines([self.a_rev, self.b_rev])
        self.lines_a = lines[self.a_rev]
        self.lines_b = lines[self.b_rev]

    def get_lines(self, revisions):
        """Get lines for revisions from the backing VersionedFiles.

        :raises RevisionNotPresent: on absent texts.
        """
        keys = [(self._key_prefix + (rev,)) for rev in revisions]
        result = {}
        for record in self.vf.get_record_stream(keys, "unordered", True):
            if record.storage_kind == "absent":
                raise RevisionNotPresent(record.key, self.vf)
            result[record.key[-1]] = record.get_bytes_as("lines")
        return result

    def plan_merge(self):
        """Generate a 'plan' for merging the two revisions.

        This involves comparing their texts and determining the cause of
        differences.  If text A has a line and text B does not, then either the
        line was added to text A, or it was deleted from B.  Once the causes
        are combined, they are written out in the format described in
        VersionedFile.plan_merge
        """
        blocks = self._get_matching_blocks(self.a_rev, self.b_rev)
        unique_a, unique_b = self._unique_lines(blocks)
        new_a, killed_b = self._determine_status(self.a_rev, unique_a)
        new_b, killed_a = self._determine_status(self.b_rev, unique_b)
        return self._iter_plan(blocks, new_a, killed_b, new_b, killed_a)

    def _iter_plan(self, blocks, new_a, killed_b, new_b, killed_a):
        last_i = 0
        last_j = 0
        for i, j, n in blocks:
            for a_index in range(last_i, i):
                if a_index in new_a:
                    if a_index in killed_b:
                        yield "conflicted-a", self.lines_a[a_index]
                    else:
                        yield "new-a", self.lines_a[a_index]
                else:
                    yield "killed-b", self.lines_a[a_index]
            for b_index in range(last_j, j):
                if b_index in new_b:
                    if b_index in killed_a:
                        yield "conflicted-b", self.lines_b[b_index]
                    else:
                        yield "new-b", self.lines_b[b_index]
                else:
                    yield "killed-a", self.lines_b[b_index]
            # handle common lines
            for a_index in range(i, i + n):
                yield "unchanged", self.lines_a[a_index]
            last_i = i + n
            last_j = j + n

    def _get_matching_blocks(self, left_revision, right_revision):
        """Return a description of which sections of two revisions match.

        See SequenceMatcher.get_matching_blocks
        """
        cached = self._cached_matching_blocks.get((left_revision, right_revision))
        if cached is not None:
            return cached
        if self._last_lines_revision_id == left_revision:
            left_lines = self._last_lines
            right_lines = self.get_lines([right_revision])[right_revision]
        else:
            lines = self.get_lines([left_revision, right_revision])
            left_lines = lines[left_revision]
            right_lines = lines[right_revision]
        self._last_lines = right_lines
        self._last_lines_revision_id = right_revision
        matcher = patiencediff.PatienceSequenceMatcher(None, left_lines, right_lines)
        return matcher.get_matching_blocks()

    def _unique_lines(self, matching_blocks):
        """Analyse matching_blocks to determine which lines are unique.

        :return: a tuple of (unique_left, unique_right), where the values are
            sets of line numbers of unique lines.
        """
        last_i = 0
        last_j = 0
        unique_left = []
        unique_right = []
        for i, j, n in matching_blocks:
            unique_left.extend(range(last_i, i))
            unique_right.extend(range(last_j, j))
            last_i = i + n
            last_j = j + n
        return unique_left, unique_right

    @staticmethod
    def _subtract_plans(old_plan, new_plan):
        """Remove changes from new_plan that came from old_plan.

        It is assumed that the difference between the old_plan and new_plan
        is their choice of 'b' text.

        All lines from new_plan that differ from old_plan are emitted
        verbatim.  All lines from new_plan that match old_plan but are
        not about the 'b' revision are emitted verbatim.

        Lines that match and are about the 'b' revision are the lines we
        don't want, so we convert 'killed-b' -> 'unchanged', and 'new-b'
        is skipped entirely.
        """
        matcher = patiencediff.PatienceSequenceMatcher(None, old_plan, new_plan)
        last_j = 0
        for _i, j, n in matcher.get_matching_blocks():
            for jj in range(last_j, j):
                yield new_plan[jj]
            for jj in range(j, j + n):
                plan_line = new_plan[jj]
                if plan_line[0] == "new-b":
                    pass
                elif plan_line[0] == "killed-b":
                    yield "unchanged", plan_line[1]
                else:
                    yield plan_line
            last_j = j + n


class _PlanMerge(_PlanMergeBase):
    """Plan an annotate merge using on-the-fly annotation."""

    def __init__(self, a_rev, b_rev, vf, key_prefix):
        super().__init__(a_rev, b_rev, vf, key_prefix)
        self.a_key = self._key_prefix + (self.a_rev,)
        self.b_key = self._key_prefix + (self.b_rev,)
        self.graph = _mod_graph.Graph(self.vf)
        heads = self.graph.heads((self.a_key, self.b_key))
        if len(heads) == 1:
            # one side dominates, so we can just return its values, yay for
            # per-file graphs
            # Ideally we would know that before we get this far
            self._head_key = heads.pop()
            other = b_rev if self._head_key == self.a_key else a_rev
            logger.debug(
                "found dominating revision for %s\n%s > %s",
                self.vf,
                self._head_key[-1],
                other,
            )
            self._weave = None
        else:
            self._head_key = None
            self._build_weave()

    def _precache_tip_lines(self):
        # Turn this into a no-op, because we will do this later
        pass

    def _find_recursive_lcas(self):
        """Find all the ancestors back to a unique lca."""
        cur_ancestors = (self.a_key, self.b_key)
        # graph.find_lca(uncommon, keys) now returns plain NULL_REVISION,
        # rather than a key tuple. We will just map that directly to no common
        # ancestors.
        parent_map = {}
        while True:
            next_lcas = self.graph.find_lca(*cur_ancestors)
            # Map a plain NULL_REVISION to a simple no-ancestors
            if next_lcas == {_mod_revision.NULL_REVISION}:
                next_lcas = ()
            # Order the lca's based on when they were merged into the tip
            # While the actual merge portion of weave merge uses a set() of
            # active revisions, the order of insertion *does* effect the
            # implicit ordering of the texts.
            for rev_key in cur_ancestors:
                ordered_parents = tuple(self.graph.find_merge_order(rev_key, next_lcas))
                parent_map[rev_key] = ordered_parents
            if len(next_lcas) == 0:
                break
            elif len(next_lcas) == 1:
                parent_map[list(next_lcas)[0]] = ()
                break
            elif len(next_lcas) > 2:
                # More than 2 lca's, fall back to grabbing all nodes between
                # this and the unique lca.
                logger.debug(
                    "More than 2 LCAs, falling back to all nodes for: %s, %s\n=> %s",
                    self.a_key,
                    self.b_key,
                    cur_ancestors,
                )
                cur_lcas = next_lcas
                while len(cur_lcas) > 1:
                    cur_lcas = self.graph.find_lca(*cur_lcas)
                if len(cur_lcas) == 0:
                    # No common base to find, use the full ancestry
                    unique_lca = None
                else:
                    unique_lca = list(cur_lcas)[0]
                    if unique_lca == _mod_revision.NULL_REVISION:
                        # find_lca will return a plain 'NULL_REVISION' rather
                        # than a key tuple when there is no common ancestor, we
                        # prefer to just use None, because it doesn't confuse
                        # _get_interesting_texts()
                        unique_lca = None
                parent_map.update(self._find_unique_parents(next_lcas, unique_lca))
                break
            cur_ancestors = next_lcas
        return parent_map

    def _find_unique_parents(self, tip_keys, base_key):
        """Find ancestors of tip that aren't ancestors of base.

        :param tip_keys: Nodes that are interesting
        :param base_key: Cull all ancestors of this node
        :return: The parent map for all revisions between tip_keys and
            base_key. base_key will be included. References to nodes outside of
            the ancestor set will also be removed.
        """
        # TODO: this would be simpler if find_unique_ancestors took a list
        #       instead of a single tip, internally it supports it, but it
        #       isn't a "backwards compatible" api change.
        if base_key is None:
            parent_map = dict(self.graph.iter_ancestry(tip_keys))
            # We remove NULL_REVISION because it isn't a proper tuple key, and
            # thus confuses things like _get_interesting_texts, and our logic
            # to add the texts into the memory weave.
            if _mod_revision.NULL_REVISION in parent_map:
                parent_map.pop(_mod_revision.NULL_REVISION)
        else:
            interesting = set()
            for tip in tip_keys:
                interesting.update(self.graph.find_unique_ancestors(tip, [base_key]))
            parent_map = self.graph.get_parent_map(interesting)
            parent_map[base_key] = ()
        culled_parent_map, child_map, tails = self._remove_external_references(
            parent_map
        )
        # Remove all the tails but base_key
        if base_key is not None:
            tails.remove(base_key)
            self._prune_tails(culled_parent_map, child_map, tails)
        # Now remove all the uninteresting 'linear' regions
        simple_map = _mod_graph.collapse_linear_regions(culled_parent_map)
        return simple_map

    @staticmethod
    def _remove_external_references(parent_map):
        """Remove references that go outside of the parent map.

        :param parent_map: Something returned from Graph.get_parent_map(keys)
        :return: (filtered_parent_map, child_map, tails)
            filtered_parent_map is parent_map without external references
            child_map is the {parent_key: [child_keys]} mapping
            tails is a list of nodes that do not have any parents in the map
        """
        # TODO: The basic effect of this function seems more generic than
        #       _PlanMerge. But the specific details of building a child_map,
        #       and computing tails seems very specific to _PlanMerge.
        #       Still, should this be in Graph land?
        filtered_parent_map = {}
        child_map = {}
        tails = []
        for key, parent_keys in parent_map.items():
            culled_parent_keys = [p for p in parent_keys if p in parent_map]
            if not culled_parent_keys:
                tails.append(key)
            for parent_key in culled_parent_keys:
                child_map.setdefault(parent_key, []).append(key)
            # TODO: Do we want to do this, it adds overhead for every node,
            #       just to say that the node has no children
            child_map.setdefault(key, [])
            filtered_parent_map[key] = culled_parent_keys
        return filtered_parent_map, child_map, tails

    @staticmethod
    def _prune_tails(parent_map, child_map, tails_to_remove):
        """Remove tails from the parent map.

        This will remove the supplied revisions until no more children have 0
        parents.

        :param parent_map: A dict of {child: [parents]}, this dictionary will
            be modified in place.
        :param tails_to_remove: A list of tips that should be removed,
            this list will be consumed
        :param child_map: The reverse dict of parent_map ({parent: [children]})
            this dict will be modified
        :return: None, parent_map will be modified in place.
        """
        while tails_to_remove:
            next = tails_to_remove.pop()
            parent_map.pop(next)
            children = child_map.pop(next)
            for child in children:
                child_parents = parent_map[child]
                child_parents.remove(next)
                if len(child_parents) == 0:
                    tails_to_remove.append(child)

    def _get_interesting_texts(self, parent_map):
        """Return a dict of texts we are interested in.

        Note that the input is in key tuples, but the output is in plain
        revision ids.

        :param parent_map: The output from _find_recursive_lcas
        :return: A dict of {'revision_id':lines} as returned by
            _PlanMergeBase.get_lines()
        """
        all_revision_keys = set(parent_map)
        all_revision_keys.add(self.a_key)
        all_revision_keys.add(self.b_key)

        # Everything else is in 'keys' but get_lines is in 'revision_ids'
        all_texts = self.get_lines([k[-1] for k in all_revision_keys])
        return all_texts

    def _build_weave(self):

        self._weave = weave.Weave(weave_name="in_memory_weave", allow_reserved=True)
        parent_map = self._find_recursive_lcas()

        all_texts = self._get_interesting_texts(parent_map)

        # Note: Unfortunately, the order given by topo_sort will effect the
        # ordering resolution in the output. Specifically, if you add A then B,
        # then in the output text A lines will show up before B lines. And, of
        # course, topo_sort doesn't guarantee any real ordering.
        # So we use merge_sort, and add a fake node on the tip.
        # This ensures that left-hand parents will always be inserted into the
        # weave before right-hand parents.
        tip_key = self._key_prefix + (_mod_revision.CURRENT_REVISION,)
        parent_map[tip_key] = (self.a_key, self.b_key)

        for _seq_num, key, _depth, _eom in reversed(merge_sort(parent_map, tip_key)):
            if key == tip_key:
                continue
            # for key in tsort.topo_sort(parent_map):
            parent_keys = parent_map[key]
            revision_id = key[-1]
            parent_ids = [k[-1] for k in parent_keys]
            self._weave.add_lines(revision_id, parent_ids, all_texts[revision_id])

    def plan_merge(self):
        """Generate a 'plan' for merging the two revisions.

        This involves comparing their texts and determining the cause of
        differences.  If text A has a line and text B does not, then either the
        line was added to text A, or it was deleted from B.  Once the causes
        are combined, they are written out in the format described in
        VersionedFile.plan_merge
        """
        if self._head_key is not None:  # There was a single head
            if self._head_key == self.a_key:
                plan = "new-a"
            else:
                if self._head_key != self.b_key:
                    raise AssertionError(
                        f"There was an invalid head: {self.b_key} != {self._head_key}"
                    )
                plan = "new-b"
            head_rev = self._head_key[-1]
            lines = self.get_lines([head_rev])[head_rev]
            return ((plan, line) for line in lines)
        return self._weave.plan_merge(self.a_rev, self.b_rev)


class _PlanLCAMerge(_PlanMergeBase):
    """Merger that uses LCA.

    This merge algorithm differs from _PlanMerge in that:

    1. comparisons are done against LCAs only
    2. cases where a contested line is new versus one LCA but old versus
       another are marked as conflicts, by emitting the line as conflicted-a
       or conflicted-b.

    This is faster, and hopefully produces more useful output.
    """

    def __init__(self, a_rev, b_rev, vf, key_prefix, graph):
        _PlanMergeBase.__init__(self, a_rev, b_rev, vf, key_prefix)
        lcas = graph.find_lca(key_prefix + (a_rev,), key_prefix + (b_rev,))
        self.lcas = set()
        for lca in lcas:
            if lca == _mod_revision.NULL_REVISION:
                self.lcas.add(lca)
            else:
                self.lcas.add(lca[-1])
        for lca in self.lcas:
            lca_lines = [] if _mod_revision.is_null(lca) else self.get_lines([lca])[lca]
            matcher = patiencediff.PatienceSequenceMatcher(
                None, self.lines_a, lca_lines
            )
            blocks = list(matcher.get_matching_blocks())
            self._cached_matching_blocks[(a_rev, lca)] = blocks
            matcher = patiencediff.PatienceSequenceMatcher(
                None, self.lines_b, lca_lines
            )
            blocks = list(matcher.get_matching_blocks())
            self._cached_matching_blocks[(b_rev, lca)] = blocks

    def _determine_status(self, revision_id, unique_line_numbers):
        """Determines the status unique lines versus all lcas.

        Basically, determines why the line is unique to this revision.

        A line may be determined new, killed, or both.

        If a line is determined new, that means it was not present in at least
        one LCA, and is not present in the other merge revision.

        If a line is determined killed, that means the line was present in
        at least one LCA.

        If a line is killed and new, this indicates that the two merge
        revisions contain differing conflict resolutions.

        :param revision_id: The id of the revision in which the lines are
            unique
        :param unique_line_numbers: The line numbers of unique lines.
        :return: a tuple of (new_this, killed_other)
        """
        new = set()
        killed = set()
        unique_line_numbers = set(unique_line_numbers)
        for lca in self.lcas:
            blocks = self._get_matching_blocks(revision_id, lca)
            unique_vs_lca, _ignored = self._unique_lines(blocks)
            new.update(unique_line_numbers.intersection(unique_vs_lca))
            killed.update(unique_line_numbers.difference(unique_vs_lca))
        return new, killed



# Copyright (C) 2007, 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Python implementations of Dirstate Helper functions."""

# We cannot import the dirstate module, because it loads this module


def _read_dirblocks(state):
    """Read in the dirblocks for the given DirState object.

    This is tightly bound to the DirState internal representation. It should be
    thought of as a member function, which is only separated out so that we can
    re-write it in pyrex.

    :param state: A DirState object.
    :return: None
    """
    from .dirstate import DirState, DirstateCorrupt, _fields_per_entry

    state._state_file.seek(state._end_of_header)
    text = state._state_file.read()
    # TODO: check the crc checksums. crc_measured = zlib.crc32(text)

    fields = text.split(b"\0")
    # Remove the last blank entry
    trailing = fields.pop()
    if trailing != b"":
        raise DirstateCorrupt(state, f"trailing garbage: {trailing!r}")
    # consider turning fields into a tuple.

    # skip the first field which is the trailing null from the header.
    cur = 1
    # Each line now has an extra '\n' field which is not used
    # so we just skip over it
    # entry size:
    #  3 fields for the key
    #  + number of fields per tree_data (5) * tree count
    #  + newline
    num_present_parents = state._num_present_parents()
    1 + num_present_parents
    entry_size = _fields_per_entry(num_present_parents)
    expected_field_count = entry_size * state._num_entries
    field_count = len(fields)
    # this checks our adjustment, and also catches file too short.
    if field_count - cur != expected_field_count:
        raise DirstateCorrupt(
            state,
            "field count incorrect {} != {}, entry_size={}, "
            "num_entries={} fields={!r}".format(
                field_count - cur,
                expected_field_count,
                entry_size,
                state._num_entries,
                fields,
            ),
        )

    if num_present_parents == 1:
        # Bind external functions to local names
        _int = int
        # We access all fields in order, so we can just iterate over
        # them. Grab an straight iterator over the fields. (We use an
        # iterator because we don't want to do a lot of additions, nor
        # do we want to do a lot of slicing)
        _iter = iter(fields)
        # Get a local reference to the compatible next method
        next = getattr(_iter, "__next__", None)
        if next is None:
            next = _iter.next
        # Move the iterator to the current position
        for _x in range(cur):
            next()
        # The two blocks here are deliberate: the root block and the
        # contents-of-root block.
        state._dirblocks = [(b"", []), (b"", [])]
        current_block = state._dirblocks[0][1]
        current_dirname = b""
        append_entry = current_block.append
        for _count in range(state._num_entries):
            dirname = next()
            name = next()
            file_id = next()
            if dirname != current_dirname:
                # new block - different dirname
                current_block = []
                current_dirname = dirname
                state._dirblocks.append((current_dirname, current_block))
                append_entry = current_block.append
            # we know current_dirname == dirname, so re-use it to avoid
            # creating new strings
            entry = (
                (current_dirname, name, file_id),
                [
                    (  # Current Tree
                        next(),  # minikind
                        next(),  # fingerprint
                        _int(next()),  # size
                        next() == b"y",  # executable
                        next(),  # packed_stat or revision_id
                    ),
                    (  # Parent 1
                        next(),  # minikind
                        next(),  # fingerprint
                        _int(next()),  # size
                        next() == b"y",  # executable
                        next(),  # packed_stat or revision_id
                    ),
                ],
            )
            trailing = next()
            if trailing != b"\n":
                raise ValueError(f"trailing garbage in dirstate: {trailing!r}")
            # append the entry to the current block
            append_entry(entry)
        state._split_root_dirblock_into_contents()
    else:
        fields_to_entry = state._get_fields_to_entry()
        entries = [
            fields_to_entry(fields[pos : pos + entry_size])
            for pos in range(cur, field_count, entry_size)
        ]
        state._entries_to_current_state(entries)
    # To convert from format 2  => format 3
    # state._dirblocks = sorted(state._dirblocks,
    #                          key=lambda blk:blk[0].split('/'))
    # To convert from format 3 => format 2
    # state._dirblocks = sorted(state._dirblocks)
    state._dirblock_state = DirState.IN_MEMORY_UNMODIFIED



# Copyright (C) 2005-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""XML externalization support."""

# "XML is like violence: if it doesn't solve your problem, you aren't
# using enough of it." -- various

# importing this module is fairly slow because it has to load several
# ElementTree bits

__all__ = [
    "Element",
    "ElementTree",
    "SubElement",
    "escape_invalid_chars",
    "fromstring",
    "fromstringlist",
    "get_utf8_or_ascii",
    "serialize_inventory_flat",
    "tostring",
    "tostringlist",
    "unpack_inventory_entry",
    "unpack_inventory_flat",
]

from xml.etree.ElementTree import (
    Element,
    ElementTree,
    ParseError,
    SubElement,
    fromstring,
    fromstringlist,
    tostring,
    tostringlist,
)

from . import inventory, serializer


class XMLRevisionSerializer(serializer.RevisionSerializer):
    """Abstract XML object serialize/deserialize."""

    squashes_xml_invalid_characters = True

    def _unpack_revision(self, element):
        raise NotImplementedError(self._unpack_revision)

    def write_revision_to_string(self, rev):
        """Serialize a revision object to a UTF-8 string."""
        return b"".join(self.write_revision_to_lines(rev))

    def read_revision(self, f):
        """Read a revision from an open file object."""
        return self._unpack_revision(self._read_element(f))

    def read_revision_from_string(self, xml_string):
        """Read a revision from an XML string."""
        return self._unpack_revision(fromstring(xml_string))  # noqa: S314

    def _read_element(self, f):
        return ElementTree().parse(f)


class XMLInventorySerializer(serializer.InventorySerializer):
    """Abstract XML object serialize/deserialize."""

    def read_inventory_from_lines(
        self, lines, revision_id=None, entry_cache=None, return_from_cache=False
    ):
        """Read xml_string into an inventory object.

        :param chunks: The xml to read.
        :param revision_id: If not-None, the expected revision id of the
            inventory. Some serialisers use this to set the results' root
            revision. This should be supplied for deserialising all
            from-repository inventories so that xml5 inventories that were
            serialised without a revision identifier can be given the right
            revision id (but not for working tree inventories where users can
            edit the data without triggering checksum errors or anything).
        :param entry_cache: An optional cache of InventoryEntry objects. If
            supplied we will look up entries via (file_id, revision_id) which
            should map to a valid InventoryEntry (File/Directory/etc) object.
        :param return_from_cache: Return entries directly from the cache,
            rather than copying them first. This is only safe if the caller
            promises not to mutate the returned inventory entries, but it can
            make some operations significantly faster.
        """
        try:
            return self._unpack_inventory(
                fromstringlist(lines),
                revision_id,
                entry_cache=entry_cache,
                return_from_cache=return_from_cache,
            )
        except ParseError as e:
            raise serializer.UnexpectedInventoryFormat(str(e)) from e

    def _unpack_inventory(
        self,
        element,
        revision_id: bytes | None = None,
        entry_cache=None,
        return_from_cache=False,
    ):
        raise NotImplementedError(self._unpack_inventory)

    def read_inventory(self, f, revision_id=None):
        """Read an inventory from an open file object."""
        try:
            try:
                return self._unpack_inventory(self._read_element(f), revision_id=None)
            finally:
                f.close()
        except ParseError as e:
            raise serializer.UnexpectedInventoryFormat(str(e)) from e

    def _read_element(self, f):
        return ElementTree().parse(f)


def get_utf8_or_ascii(a_str):
    """Return a cached version of the string.

    cElementTree will return a plain string if the XML is plain ascii. It only
    returns Unicode when it needs to. We want to work in utf-8 strings. So if
    cElementTree returns a plain string, we can just return the cached version.
    If it is Unicode, then we need to encode it.

    :param a_str: An 8-bit string or Unicode as returned by
                  cElementTree.Element.get()
    :return: A utf-8 encoded 8-bit string.
    """
    # This is fairly optimized because we know what cElementTree does, this is
    # not meant as a generic function for all cases. Because it is possible for
    # an 8-bit string to not be ascii or valid utf8.
    if a_str.__class__ is str:
        return a_str.encode("utf-8")
    else:
        return a_str


from ._bzr_rs import encode_and_escape, escape_invalid_chars


def unpack_inventory_entry(
    elt, entry_cache=None, return_from_cache=False, root_id=None
):
    """Unpack an inventory entry from XML element."""
    elt_get = elt.get
    file_id = elt_get("file_id")
    revision = elt_get("revision")
    # Check and see if we have already unpacked this exact entry
    # Some timings for "repo.revision_trees(last_100_revs)"
    #               bzr     mysql
    #   unmodified  4.1s    40.8s
    #   using lru   3.5s
    #   using fifo  2.83s   29.1s
    #   lru._cache  2.8s
    #   dict        2.75s   26.8s
    #   inv.add     2.5s    26.0s
    #   no_copy     2.00s   20.5s
    #   no_c,dict   1.95s   18.0s
    # Note that a cache of 10k nodes is more than sufficient to hold all of
    # the inventory for the last 100 revs for bzr, but not for mysql (20k
    # is enough for mysql, which saves the same 2s as using a dict)

    # Breakdown of mysql using time.clock()
    #   4.1s    2 calls to element.get for file_id, revision_id
    #   4.5s    cache_hit lookup
    #   7.1s    InventoryFile.copy()
    #   2.4s    InventoryDirectory.copy()
    #   0.4s    decoding unique entries
    #   1.6s    decoding entries after FIFO fills up
    #   0.8s    Adding nodes to FIFO (including flushes)
    #   0.1s    cache miss lookups
    # Using an LRU cache
    #   4.1s    2 calls to element.get for file_id, revision_id
    #   9.9s    cache_hit lookup
    #   10.8s   InventoryEntry.copy()
    #   0.3s    cache miss lookus
    #   1.2s    decoding entries
    #   1.0s    adding nodes to LRU
    if entry_cache is not None and revision is not None:
        key = (file_id, revision)
        try:
            # We copy it, because some operations may mutate it
            cached_ie = entry_cache[key]
        except KeyError:
            pass
        else:
            # Only copying directory entries drops us 2.85s => 2.35s
            if return_from_cache:
                if cached_ie.kind == "directory":
                    return cached_ie.copy()
                return cached_ie
            return cached_ie.copy()

    kind = elt.tag
    if not inventory.InventoryEntry.versionable_kind(kind):
        raise AssertionError(f"unsupported entry kind {kind}")

    file_id = get_utf8_or_ascii(file_id)
    if revision is not None:
        revision = get_utf8_or_ascii(revision)
    parent_id = elt_get("parent_id")
    parent_id = get_utf8_or_ascii(parent_id) if parent_id is not None else root_id

    if kind == "directory":
        ie = inventory.InventoryDirectory(file_id, elt_get("name"), parent_id, revision)
    elif kind == "file":
        text_sha1 = elt_get("text_sha1")
        if text_sha1 is not None:
            text_sha1 = text_sha1.encode("ascii")
        executable = elt_get("executable") == "yes"
        v = elt_get("text_size")
        text_size = v and int(v)
        ie = inventory.InventoryFile(
            file_id,
            elt_get("name"),
            parent_id,
            revision,
            text_sha1=text_sha1,
            executable=executable,
            text_size=text_size,
        )
    elif kind == "symlink":
        symlink_target = elt_get("symlink_target")
        ie = inventory.InventoryLink(
            file_id, elt_get("name"), parent_id, revision, symlink_target=symlink_target
        )
    elif kind == "tree-reference":
        file_id = get_utf8_or_ascii(elt.attrib["file_id"])
        name = elt.attrib["name"]
        parent_id = get_utf8_or_ascii(elt.attrib["parent_id"])
        revision = get_utf8_or_ascii(elt.get("revision"))
        reference_revision = get_utf8_or_ascii(elt.get("reference_revision"))
        ie = inventory.TreeReference(
            file_id, name, parent_id, revision, reference_revision
        )
    else:
        raise serializer.UnsupportedInventoryKind(kind)
    if revision is not None and entry_cache is not None:
        # We cache a copy() because callers like to mutate objects, and
        # that would cause the item in cache to mutate as well.
        # This has a small effect on many-inventory performance, because
        # the majority fraction is spent in cache hits, not misses.
        entry_cache[key] = ie.copy()

    return ie


def unpack_inventory_flat(
    elt, format_num, unpack_entry, entry_cache=None, return_from_cache=False
):
    """Unpack a flat XML inventory.

    :param elt: XML element for the inventory
    :param format_num: Expected format number
    :param unpack_entry: Function for unpacking inventory entries
    :return: An inventory
    :raise UnexpectedInventoryFormat: When unexpected elements or data is
        encountered
    """
    if elt.tag != "inventory":
        raise serializer.UnexpectedInventoryFormat(f"Root tag is {elt.tag!r}")
    format = elt.get("format")
    if (format is None and format_num is not None) or format.encode() != format_num:
        raise serializer.UnexpectedInventoryFormat(f"Invalid format version {format!r}")
    revision_id = elt.get("revision_id")
    if revision_id is not None:
        revision_id = revision_id.encode("utf-8")
    inv = inventory.Inventory(root_id=None, revision_id=revision_id)
    for e in elt:
        ie = unpack_entry(e, entry_cache, return_from_cache)
        inv.add(ie)
    return inv


def serialize_inventory_flat(inv, append, root_id, supported_kinds, working):
    """Serialize an inventory to a flat XML file.

    :param inv: Inventory to serialize
    :param append: Function for writing a line of output
    :param working: If True skip history data - text_sha1, text_size,
        reference_revision, symlink_target.
    """
    entries = inv.iter_entries()
    # Skip the root
    _root_path, _root_ie = next(entries)
    for _path, ie in entries:
        if ie.parent_id != root_id:
            parent_str = b"".join(
                [b' parent_id="', encode_and_escape(ie.parent_id), b'"']
            )
        else:
            parent_str = b""
        if ie.kind == "file":
            executable = b' executable="yes"' if ie.executable else b""
            if not working:
                append(
                    b'<file%s file_id="%s" name="%s"%s revision="%s" '
                    b'text_sha1="%s" text_size="%d" />\n'
                    % (
                        executable,
                        encode_and_escape(ie.file_id),
                        encode_and_escape(ie.name),
                        parent_str,
                        encode_and_escape(ie.revision),
                        ie.text_sha1,
                        ie.text_size,
                    )
                )
            else:
                append(
                    b'<file%s file_id="%s" name="%s"%s />\n'
                    % (
                        executable,
                        encode_and_escape(ie.file_id),
                        encode_and_escape(ie.name),
                        parent_str,
                    )
                )
        elif ie.kind == "directory":
            if not working:
                append(
                    b'<directory file_id="%s" name="%s"%s revision="%s" '
                    b"/>\n"
                    % (
                        encode_and_escape(ie.file_id),
                        encode_and_escape(ie.name),
                        parent_str,
                        encode_and_escape(ie.revision),
                    )
                )
            else:
                append(
                    b'<directory file_id="%s" name="%s"%s />\n'
                    % (
                        encode_and_escape(ie.file_id),
                        encode_and_escape(ie.name),
                        parent_str,
                    )
                )
        elif ie.kind == "symlink":
            if not working:
                append(
                    b'<symlink file_id="%s" name="%s"%s revision="%s" '
                    b'symlink_target="%s" />\n'
                    % (
                        encode_and_escape(ie.file_id),
                        encode_and_escape(ie.name),
                        parent_str,
                        encode_and_escape(ie.revision),
                        encode_and_escape(ie.symlink_target),
                    )
                )
            else:
                append(
                    b'<symlink file_id="%s" name="%s"%s />\n'
                    % (
                        encode_and_escape(ie.file_id),
                        encode_and_escape(ie.name),
                        parent_str,
                    )
                )
        elif ie.kind == "tree-reference":
            if ie.kind not in supported_kinds:
                raise serializer.UnsupportedInventoryKind(ie.kind)
            if not working:
                append(
                    b'<tree-reference file_id="%s" name="%s"%s '
                    b'revision="%s" reference_revision="%s" />\n'
                    % (
                        encode_and_escape(ie.file_id),
                        encode_and_escape(ie.name),
                        parent_str,
                        encode_and_escape(ie.revision),
                        encode_and_escape(ie.reference_revision),
                    )
                )
            else:
                append(
                    b'<tree-reference file_id="%s" name="%s"%s />\n'
                    % (
                        encode_and_escape(ie.file_id),
                        encode_and_escape(ie.name),
                        parent_str,
                    )
                )
        else:
            raise serializer.UnsupportedInventoryKind(ie.kind)
    append(b"</inventory>\n")



# Copyright (C) 2025 Breezy Contributors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for bzrformats error classes."""

from .. import errors
from . import TestCase


class TestNoSuchFile(TestCase):
    """Test NoSuchFile error."""

    def test_no_such_file_str(self):
        """Test string representation of NoSuchFile."""
        err = errors.NoSuchFile("/path/to/missing/file")
        self.assertEqual("No such file: '/path/to/missing/file'", str(err))

    def test_no_such_file_with_extra(self):
        """Test NoSuchFile with extra information."""
        err = errors.NoSuchFile("/path/to/file", "additional info")
        self.assertEqual("No such file: '/path/to/file': additional info", str(err))


class TestPathError(TestCase):
    """Test PathError base class."""

    def test_path_error_str(self):
        """Test string representation of PathError."""
        err = errors.PathError("/some/path")
        self.assertEqual("Path error: '/some/path'", str(err))

    def test_path_error_with_extra(self):
        """Test PathError with extra information."""
        err = errors.PathError("/some/path", "extra details")
        self.assertEqual("Path error: '/some/path': extra details", str(err))


class TestReservedId(TestCase):
    """Test ReservedId error."""

    def test_reserved_id_str(self):
        """Test string representation of ReservedId."""
        err = errors.ReservedId(b"null:")
        self.assertEqual("Reserved revision-id {b'null:'}", str(err))


class TestRevisionNotPresent(TestCase):
    """Test RevisionNotPresent error."""

    def test_revision_not_present_str(self):
        """Test string representation of RevisionNotPresent."""
        err = errors.RevisionNotPresent(b"rev-123", b"file-456")
        expected = "Revision {b'rev-123'} not present in \"b'file-456'\"."
        self.assertEqual(expected, str(err))


class TestRevisionAlreadyPresent(TestCase):
    """Test RevisionAlreadyPresent error."""

    def test_revision_already_present_str(self):
        """Test string representation of RevisionAlreadyPresent."""
        err = errors.RevisionAlreadyPresent(b"rev-123", b"file-456")
        expected = "Revision {b'rev-123'} already present in \"b'file-456'\"."
        self.assertEqual(expected, str(err))


class TestInvalidRevisionId(TestCase):
    """Test InvalidRevisionId error."""

    def test_invalid_revision_id_str(self):
        """Test string representation of InvalidRevisionId."""
        err = errors.InvalidRevisionId(b"bad-rev", "mybranch")
        expected = "Invalid revision-id {b'bad-rev'} in mybranch"
        self.assertEqual(expected, str(err))


class TestNoSuchId(TestCase):
    """Test NoSuchId error."""

    def test_no_such_id_str(self):
        """Test string representation of NoSuchId."""
        from bzrformats.inventory import NoSuchId

        err = NoSuchId("tree-object", b"file-id-123")
        expected = (
            "The file id \"b'file-id-123'\" is not present in the tree tree-object."
        )
        self.assertEqual(expected, str(err))


class TestInconsistentDelta(TestCase):
    def test_inconsistent_delta_str(self):
        err = errors.InconsistentDelta("path", "file-id", "reason for foo")
        self.assertEqual(
            "An inconsistent delta was supplied involving 'path', 'file-id'\n"
            "reason: reason for foo",
            str(err),
        )


# Add test module discovery
def test_suite():
    """Return the test suite for error tests."""
    import unittest

    return unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])


if __name__ == "__main__":
    import unittest

    unittest.main()



# Copyright (C) 2005-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


# TODO: tests regarding version names
# TODO: rbc 20050108 test that join does not leave an inconsistent weave
#       if it fails.

"""test suite for weave algorithm."""

from io import BytesIO
from pprint import pformat

from ..errors import ReservedId, RevisionAlreadyPresent, RevisionNotPresent
from ..osutils import sha_string
from ..weave import Weave, WeaveFormatError, WeaveInvalidChecksum
from ..weavefile import read_weave, write_weave
from . import TestCase, TestCaseInTempDir

# texts for use in testing
TEXT_0 = [b"Hello world"]
TEXT_1 = [b"Hello world", b"A second line"]


class TestBase(TestCase):
    def check_read_write(self, k):
        """Check the weave k can be written & re-read."""
        from tempfile import TemporaryFile

        tf = TemporaryFile()

        write_weave(k, tf)
        tf.seek(0)
        k2 = read_weave(tf)

        if k != k2:
            tf.seek(0)
            self.log("serialized weave:")
            self.log(tf.read())

            self.log("")
            self.log("parents: %s" % (k._parents == k2._parents))
            self.log(f"         {k._parents!r}")
            self.log(f"         {k2._parents!r}")
            self.log("")
            self.fail("read/write check failed")


class WeaveContains(TestBase):
    """Weave __contains__ operator."""

    def runTest(self):
        k = Weave(get_scope=lambda: None)
        self.assertNotIn(b"foo", k)
        k.add_lines(b"foo", [], TEXT_1)
        self.assertIn(b"foo", k)


class Easy(TestBase):
    def runTest(self):
        Weave()


class AnnotateOne(TestBase):
    def runTest(self):
        k = Weave()
        k.add_lines(b"text0", [], TEXT_0)
        self.assertEqual(k.annotate(b"text0"), [(b"text0", TEXT_0[0])])


class InvalidAdd(TestBase):
    """Try to use invalid version number during add."""

    def runTest(self):
        k = Weave()

        self.assertRaises(
            RevisionNotPresent, k.add_lines, b"text0", [b"69"], [b"new text!"]
        )


class RepeatedAdd(TestBase):
    """Add the same version twice; harmless."""

    def test_duplicate_add(self):
        k = Weave()
        idx = k.add_lines(b"text0", [], TEXT_0)
        idx2 = k.add_lines(b"text0", [], TEXT_0)
        self.assertEqual(idx, idx2)


class InvalidRepeatedAdd(TestBase):
    def runTest(self):
        k = Weave()
        k.add_lines(b"basis", [], TEXT_0)
        k.add_lines(b"text0", [], TEXT_0)
        self.assertRaises(
            RevisionAlreadyPresent,
            k.add_lines,
            b"text0",
            [],
            [b"not the same text"],
        )
        self.assertRaises(
            RevisionAlreadyPresent,
            k.add_lines,
            b"text0",
            [b"basis"],  # not the right parents
            TEXT_0,
        )


class InsertLines(TestBase):
    """Store a revision that adds one line to the original.

    Look at the annotations to make sure that the first line is matched
    and not stored repeatedly.
    """

    def runTest(self):
        k = Weave()

        k.add_lines(b"text0", [], [b"line 1"])
        k.add_lines(b"text1", [b"text0"], [b"line 1", b"line 2"])

        self.assertEqual(k.annotate(b"text0"), [(b"text0", b"line 1")])

        self.assertEqual(k.get_lines(1), [b"line 1", b"line 2"])

        self.assertEqual(
            k.annotate(b"text1"), [(b"text0", b"line 1"), (b"text1", b"line 2")]
        )

        k.add_lines(b"text2", [b"text0"], [b"line 1", b"diverged line"])

        self.assertEqual(
            k.annotate(b"text2"), [(b"text0", b"line 1"), (b"text2", b"diverged line")]
        )

        text3 = [b"line 1", b"middle line", b"line 2"]
        k.add_lines(b"text3", [b"text0", b"text1"], text3)

        # self.log("changes to text3: " + pformat(list(k._delta(set([0, 1]),
        # text3))))

        self.log("k._weave=" + pformat(k._weave))

        self.assertEqual(
            k.annotate(b"text3"),
            [(b"text0", b"line 1"), (b"text3", b"middle line"), (b"text1", b"line 2")],
        )

        # now multiple insertions at different places
        k.add_lines(
            b"text4",
            [b"text0", b"text1", b"text3"],
            [b"line 1", b"aaa", b"middle line", b"bbb", b"line 2", b"ccc"],
        )

        self.assertEqual(
            k.annotate(b"text4"),
            [
                (b"text0", b"line 1"),
                (b"text4", b"aaa"),
                (b"text3", b"middle line"),
                (b"text4", b"bbb"),
                (b"text1", b"line 2"),
                (b"text4", b"ccc"),
            ],
        )


class DeleteLines(TestBase):
    """Deletion of lines from existing text.

    Try various texts all based on a common ancestor.
    """

    def runTest(self):
        k = Weave()

        base_text = [b"one", b"two", b"three", b"four"]

        k.add_lines(b"text0", [], base_text)

        texts = [
            [b"one", b"two", b"three"],
            [b"two", b"three", b"four"],
            [b"one", b"four"],
            [b"one", b"two", b"three", b"four"],
        ]

        i = 1
        for t in texts:
            k.add_lines(b"text%d" % i, [b"text0"], t)
            i += 1

        self.log("final weave:")
        self.log("k._weave=" + pformat(k._weave))

        for i in range(len(texts)):
            self.assertEqual(k.get_lines(i + 1), texts[i])


class SuicideDelete(TestBase):
    """Invalid weave which tries to add and delete simultaneously."""

    def runTest(self):
        k = Weave()

        k._parents = [
            (),
        ]
        k._weave = [
            (b"{", 0),
            b"first line",
            (b"[", 0),
            b"deleted in 0",
            (b"]", 0),
            (b"}", 0),
        ]
        # SKIPPED
        # Weave.get doesn't trap this anymore
        return

        self.assertRaises(WeaveFormatError, k.get_lines, 0)


class CannedDelete(TestBase):
    """Unpack canned weave with deleted lines."""

    def runTest(self):
        k = Weave()

        k._parents = [
            (),
            frozenset([0]),
        ]
        k._weave = [
            (b"{", 0),
            b"first line",
            (b"[", 1),
            b"line to be deleted",
            (b"]", 1),
            b"last line",
            (b"}", 0),
        ]
        k._sha1s = [
            sha_string(b"first lineline to be deletedlast line"),
            sha_string(b"first linelast line"),
        ]

        self.assertEqual(
            k.get_lines(0),
            [
                b"first line",
                b"line to be deleted",
                b"last line",
            ],
        )

        self.assertEqual(
            k.get_lines(1),
            [
                b"first line",
                b"last line",
            ],
        )


class CannedReplacement(TestBase):
    """Unpack canned weave with deleted lines."""

    def runTest(self):
        k = Weave()

        k._parents = [
            frozenset(),
            frozenset([0]),
        ]
        k._weave = [
            (b"{", 0),
            b"first line",
            (b"[", 1),
            b"line to be deleted",
            (b"]", 1),
            (b"{", 1),
            b"replacement line",
            (b"}", 1),
            b"last line",
            (b"}", 0),
        ]
        k._sha1s = [
            sha_string(b"first lineline to be deletedlast line"),
            sha_string(b"first linereplacement linelast line"),
        ]

        self.assertEqual(
            k.get_lines(0),
            [
                b"first line",
                b"line to be deleted",
                b"last line",
            ],
        )

        self.assertEqual(
            k.get_lines(1),
            [
                b"first line",
                b"replacement line",
                b"last line",
            ],
        )


class BadWeave(TestBase):
    """Test that we trap an insert which should not occur."""

    def runTest(self):
        k = Weave()

        k._parents = [
            frozenset(),
        ]
        k._weave = [
            b"bad line",
            (b"{", 0),
            b"foo {",
            (b"{", 1),
            b"  added in version 1",
            (b"{", 2),
            b"  added in v2",
            (b"}", 2),
            b"  also from v1",
            (b"}", 1),
            b"}",
            (b"}", 0),
        ]

        # SKIPPED
        # Weave.get doesn't trap this anymore
        return

        self.assertRaises(WeaveFormatError, k.get, 0)


class BadInsert(TestBase):
    """Test that we trap an insert which should not occur."""

    def runTest(self):
        k = Weave()

        k._parents = [
            frozenset(),
            frozenset([0]),
            frozenset([0]),
            frozenset([0, 1, 2]),
        ]
        k._weave = [
            (b"{", 0),
            b"foo {",
            (b"{", 1),
            b"  added in version 1",
            (b"{", 1),
            b"  more in 1",
            (b"}", 1),
            (b"}", 1),
            (b"}", 0),
        ]

        # this is not currently enforced by get
        return

        self.assertRaises(WeaveFormatError, k.get, 0)

        self.assertRaises(WeaveFormatError, k.get, 1)


class InsertNested(TestBase):
    """Insertion with nested instructions."""

    def runTest(self):
        k = Weave()

        k._parents = [
            frozenset(),
            frozenset([0]),
            frozenset([0]),
            frozenset([0, 1, 2]),
        ]
        k._weave = [
            (b"{", 0),
            b"foo {",
            (b"{", 1),
            b"  added in version 1",
            (b"{", 2),
            b"  added in v2",
            (b"}", 2),
            b"  also from v1",
            (b"}", 1),
            b"}",
            (b"}", 0),
        ]

        k._sha1s = [
            sha_string(b"foo {}"),
            sha_string(b"foo {  added in version 1  also from v1}"),
            sha_string(b"foo {  added in v2}"),
            sha_string(b"foo {  added in version 1  added in v2  also from v1}"),
        ]

        self.assertEqual(k.get_lines(0), [b"foo {", b"}"])

        self.assertEqual(
            k.get_lines(1), [b"foo {", b"  added in version 1", b"  also from v1", b"}"]
        )

        self.assertEqual(k.get_lines(2), [b"foo {", b"  added in v2", b"}"])

        self.assertEqual(
            k.get_lines(3),
            [
                b"foo {",
                b"  added in version 1",
                b"  added in v2",
                b"  also from v1",
                b"}",
            ],
        )


class DeleteLines2(TestBase):
    """Test recording revisions that delete lines.

    This relies on the weave having a way to represent lines knocked
    out by a later revision.
    """

    def runTest(self):
        k = Weave()

        k.add_lines(b"text0", [], [b"line the first", b"line 2", b"line 3", b"fine"])

        self.assertEqual(len(k.get_lines(0)), 4)

        k.add_lines(b"text1", [b"text0"], [b"line the first", b"fine"])

        self.assertEqual(k.get_lines(1), [b"line the first", b"fine"])

        self.assertEqual(
            k.annotate(b"text1"), [(b"text0", b"line the first"), (b"text0", b"fine")]
        )


class IncludeVersions(TestBase):
    """Check texts that are stored across multiple revisions.

    Here we manually create a weave with particular encoding and make
    sure it unpacks properly.

    Text 0 includes nothing; text 1 includes text 0 and adds some
    lines.
    """

    def runTest(self):
        k = Weave()

        k._parents = [frozenset(), frozenset([0])]
        k._weave = [
            (b"{", 0),
            b"first line",
            (b"}", 0),
            (b"{", 1),
            b"second line",
            (b"}", 1),
        ]

        k._sha1s = [sha_string(b"first line"), sha_string(b"first linesecond line")]

        self.assertEqual(k.get_lines(1), [b"first line", b"second line"])

        self.assertEqual(k.get_lines(0), [b"first line"])


class DivergedIncludes(TestBase):
    """Weave with two diverged texts based on version 0."""

    def runTest(self):
        # FIXME make the weave, dont poke at it.
        k = Weave()

        k._names = [b"0", b"1", b"2"]
        k._name_map = {b"0": 0, b"1": 1, b"2": 2}
        k._parents = [
            frozenset(),
            frozenset([0]),
            frozenset([0]),
        ]
        k._weave = [
            (b"{", 0),
            b"first line",
            (b"}", 0),
            (b"{", 1),
            b"second line",
            (b"}", 1),
            (b"{", 2),
            b"alternative second line",
            (b"}", 2),
        ]

        k._sha1s = [
            sha_string(b"first line"),
            sha_string(b"first linesecond line"),
            sha_string(b"first linealternative second line"),
        ]

        self.assertEqual(k.get_lines(0), [b"first line"])

        self.assertEqual(k.get_lines(1), [b"first line", b"second line"])

        self.assertEqual(k.get_lines(b"2"), [b"first line", b"alternative second line"])

        self.assertEqual(set(k.get_ancestry([b"2"])), {b"0", b"2"})


class ReplaceLine(TestBase):
    def runTest(self):
        k = Weave()

        text0 = [b"cheddar", b"stilton", b"gruyere"]
        text1 = [b"cheddar", b"blue vein", b"neufchatel", b"chevre"]

        k.add_lines(b"text0", [], text0)
        k.add_lines(b"text1", [b"text0"], text1)

        self.log("k._weave=" + pformat(k._weave))

        self.assertEqual(k.get_lines(0), text0)
        self.assertEqual(k.get_lines(1), text1)


class Merge(TestBase):
    """Storage of versions that merge diverged parents."""

    def runTest(self):
        k = Weave()

        texts = [
            [b"header"],
            [b"header", b"", b"line from 1"],
            [b"header", b"", b"line from 2", b"more from 2"],
            [b"header", b"", b"line from 1", b"fixup line", b"line from 2"],
        ]

        k.add_lines(b"text0", [], texts[0])
        k.add_lines(b"text1", [b"text0"], texts[1])
        k.add_lines(b"text2", [b"text0"], texts[2])
        k.add_lines(b"merge", [b"text0", b"text1", b"text2"], texts[3])

        for i, t in enumerate(texts):
            self.assertEqual(k.get_lines(i), t)

        self.assertEqual(
            k.annotate(b"merge"),
            [
                (b"text0", b"header"),
                (b"text1", b""),
                (b"text1", b"line from 1"),
                (b"merge", b"fixup line"),
                (b"text2", b"line from 2"),
            ],
        )

        self.assertEqual(
            set(k.get_ancestry([b"merge"])), {b"text0", b"text1", b"text2", b"merge"}
        )

        self.log("k._weave=" + pformat(k._weave))

        self.check_read_write(k)


class Conflicts(TestBase):
    """Test detection of conflicting regions during a merge.

    A base version is inserted, then two descendents try to
    insert different lines in the same place.  These should be
    reported as a possible conflict and forwarded to the user.
    """

    def runTest(self):
        return  # NOT RUN
        k = Weave()

        k.add_lines([], [b"aaa", b"bbb"])
        k.add_lines([0], [b"aaa", b"111", b"bbb"])
        k.add_lines([1], [b"aaa", b"222", b"bbb"])

        k.merge([1, 2])

        self.assertEqual([[[b"aaa"]], [[b"111"], [b"222"]], [[b"bbb"]]])


class NonConflict(TestBase):
    """Two descendants insert compatible changes.

    No conflict should be reported.
    """

    def runTest(self):
        return  # NOT RUN
        k = Weave()

        k.add_lines([], [b"aaa", b"bbb"])
        k.add_lines([0], [b"111", b"aaa", b"ccc", b"bbb"])
        k.add_lines([1], [b"aaa", b"ccc", b"bbb", b"222"])


class Khayyam(TestBase):
    """Test changes to multi-line texts, and read/write."""

    def test_multi_line_merge(self):
        rawtexts = [
            b"""A Book of Verses underneath the Bough,
            A Jug of Wine, a Loaf of Bread, -- and Thou
            Beside me singing in the Wilderness --
            Oh, Wilderness were Paradise enow!""",
            b"""A Book of Verses underneath the Bough,
            A Jug of Wine, a Loaf of Bread, -- and Thou
            Beside me singing in the Wilderness --
            Oh, Wilderness were Paradise now!""",
            b"""A Book of poems underneath the tree,
            A Jug of Wine, a Loaf of Bread,
            and Thou
            Beside me singing in the Wilderness --
            Oh, Wilderness were Paradise now!

            -- O. Khayyam""",
            b"""A Book of Verses underneath the Bough,
            A Jug of Wine, a Loaf of Bread,
            and Thou
            Beside me singing in the Wilderness --
            Oh, Wilderness were Paradise now!""",
        ]
        texts = [[l.strip() for l in t.split(b"\n")] for t in rawtexts]

        k = Weave()
        parents = set()
        for i, t in enumerate(texts):
            k.add_lines(b"text%d" % i, list(parents), t)
            parents.add(b"text%d" % i)

        self.log("k._weave=" + pformat(k._weave))

        for i, t in enumerate(texts):
            self.assertEqual(k.get_lines(i), t)

        self.check_read_write(k)


class JoinWeavesTests(TestBase):
    def setUp(self):
        super().setUp()
        self.weave1 = Weave()
        self.lines1 = [b"hello\n"]
        self.lines3 = [b"hello\n", b"cruel\n", b"world\n"]
        self.weave1.add_lines(b"v1", [], self.lines1)
        self.weave1.add_lines(b"v2", [b"v1"], [b"hello\n", b"world\n"])
        self.weave1.add_lines(b"v3", [b"v2"], self.lines3)

    def test_written_detection(self):
        # Test detection of weave file corruption.
        #
        # Make sure that we can detect if a weave file has
        # been corrupted. This doesn't test all forms of corruption,
        # but it at least helps verify the data you get, is what you want.

        w = Weave()
        w.add_lines(b"v1", [], [b"hello\n"])
        w.add_lines(b"v2", [b"v1"], [b"hello\n", b"there\n"])

        tmpf = BytesIO()
        write_weave(w, tmpf)

        # Because we are corrupting, we need to make sure we have the exact
        # text
        self.assertEqual(
            b"# bzr weave file v5\n"
            b"i\n1 f572d396fae9206628714fb2ce00f72e94f2258f\nn v1\n\n"
            b"i 0\n1 90f265c6e75f1c8f9ab76dcf85528352c5f215ef\nn v2\n\n"
            b"w\n{ 0\n. hello\n}\n{ 1\n. there\n}\nW\n",
            tmpf.getvalue(),
        )

        # Change a single letter
        tmpf = BytesIO(
            b"# bzr weave file v5\n"
            b"i\n1 f572d396fae9206628714fb2ce00f72e94f2258f\nn v1\n\n"
            b"i 0\n1 90f265c6e75f1c8f9ab76dcf85528352c5f215ef\nn v2\n\n"
            b"w\n{ 0\n. hello\n}\n{ 1\n. There\n}\nW\n"
        )

        w = read_weave(tmpf)

        self.assertEqual(b"hello\n", w.get_text(b"v1"))
        self.assertRaises(WeaveInvalidChecksum, w.get_text, b"v2")
        self.assertRaises(WeaveInvalidChecksum, w.get_lines, b"v2")
        self.assertRaises(WeaveInvalidChecksum, w.check)

        # Change the sha checksum
        tmpf = BytesIO(
            b"# bzr weave file v5\n"
            b"i\n1 f572d396fae9206628714fb2ce00f72e94f2258f\nn v1\n\n"
            b"i 0\n1 f0f265c6e75f1c8f9ab76dcf85528352c5f215ef\nn v2\n\n"
            b"w\n{ 0\n. hello\n}\n{ 1\n. there\n}\nW\n"
        )

        w = read_weave(tmpf)

        self.assertEqual(b"hello\n", w.get_text(b"v1"))
        self.assertRaises(WeaveInvalidChecksum, w.get_text, b"v2")
        self.assertRaises(WeaveInvalidChecksum, w.get_lines, b"v2")
        self.assertRaises(WeaveInvalidChecksum, w.check)


class TestWeave(TestCase):
    def test_allow_reserved_false(self):
        w = Weave("name", allow_reserved=False)
        # Add lines is checked at the WeaveFile level, not at the Weave level
        w.add_lines(b"name:", [], TEXT_1)
        # But get_lines is checked at this level
        self.assertRaises(ReservedId, w.get_lines, b"name:")

    def test_allow_reserved_true(self):
        w = Weave("name", allow_reserved=True)
        w.add_lines(b"name:", [], TEXT_1)
        self.assertEqual(TEXT_1, w.get_lines(b"name:"))


class InstrumentedWeave(Weave):
    """Keep track of how many times functions are called."""

    def __init__(self, weave_name=None):
        self._extract_count = 0
        Weave.__init__(self, weave_name=weave_name)

    def _extract(self, versions):
        self._extract_count += 1
        return Weave._extract(self, versions)


class TestNeedsReweave(TestCase):
    """Internal corner cases for when reweave is needed."""

    def test_compatible_parents(self):
        w1 = Weave("a")
        my_parents = {1, 2, 3}
        # subsets are ok
        self.assertTrue(w1._compatible_parents(my_parents, {3}))
        # same sets
        self.assertTrue(w1._compatible_parents(my_parents, set(my_parents)))
        # same empty corner case
        self.assertTrue(w1._compatible_parents(set(), set()))
        # other cannot contain stuff my_parents does not
        self.assertFalse(w1._compatible_parents(set(), {1}))
        self.assertFalse(w1._compatible_parents(my_parents, {1, 2, 3, 4}))
        self.assertFalse(w1._compatible_parents(my_parents, {4}))


class TestWeaveFile(TestCaseInTempDir):
    def test_empty_file(self):
        with open("empty.weave", "wb+") as f:
            self.assertRaises(WeaveFormatError, read_weave, f)



# Copyright (C) 2005-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from ..revision import Revision
from . import TestCase


class TestRevisionMethods(TestCase):
    def test_get_summary(self):
        r = Revision(
            b"1",
            parent_ids=[],
            committer="",
            message="a",
            timestamp=0,
            timezone=0,
            inventory_sha1=None,
            properties={},
        )
        self.assertEqual("a", r.get_summary())
        r = Revision(
            b"1",
            parent_ids=[],
            committer="",
            message="a\nb",
            timestamp=0,
            timezone=0,
            inventory_sha1=None,
            properties={},
        )
        self.assertEqual("a", r.get_summary())
        r = Revision(
            b"1",
            parent_ids=[],
            committer="",
            message="\na\nb",
            timestamp=0,
            timezone=0,
            inventory_sha1=None,
            properties={},
        )
        self.assertEqual("a", r.get_summary())
        r = Revision(
            b"1",
            parent_ids=[],
            committer="",
            message="",
            timestamp=0,
            timezone=0,
            inventory_sha1=None,
            properties={},
        )
        self.assertEqual("", r.get_summary())

    def test_get_apparent_authors(self):
        r = Revision(
            b"1",
            parent_ids=[],
            committer="A",
            message="",
            timestamp=0,
            timezone=0,
            inventory_sha1=None,
            properties={},
        )
        self.assertEqual(["A"], r.get_apparent_authors())
        r = Revision(
            b"1",
            parent_ids=[],
            committer="A",
            message="",
            timestamp=0,
            timezone=0,
            inventory_sha1=None,
            properties={"author": "B"},
        )
        self.assertEqual(["B"], r.get_apparent_authors())
        r = Revision(
            b"1",
            parent_ids=[],
            committer="A",
            message="",
            timestamp=0,
            timezone=0,
            inventory_sha1=None,
            properties={"author": "B", "authors": "C\nD"},
        )
        self.assertEqual(["C", "D"], r.get_apparent_authors())

    def test_get_apparent_authors_no_committer(self):
        r = Revision(
            b"1",
            parent_ids=[],
            committer="",
            message="",
            timestamp=0,
            timezone=0,
            inventory_sha1=None,
            properties={},
        )
        self.assertEqual([], r.get_apparent_authors())



# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Bazaar-specific revision implementation."""

from bzrformats._bzr_rs import (
    CURRENT_REVISION,
    NULL_REVISION,
    check_not_reserved_id,
    is_null,
    is_reserved_id,
)
from bzrformats._bzr_rs import (
    Revision as BzrRevision,
)

RevisionID = bytes

# Re-export the Bazaar revision implementation
Revision = BzrRevision

__all__ = [
    "CURRENT_REVISION",
    "NULL_REVISION",
    "BzrRevision",
    "Revision",
    "RevisionID",
    "check_not_reserved_id",
    "is_null",
    "is_reserved_id",
]



# Copyright (C) 2008-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for group compression."""

import logging
import zlib

from testscenarios import load_tests_apply_scenarios

from .. import btree_index, groupcompress, knit, osutils, versionedfile
from .. import index as _mod_index
from ..osutils import sha_string
from . import TestCase, TestCaseWithMemoryTransport, TestNotApplicable
from .test__groupcompress import _compiled_groupcompress_module


def group_compress_implementation_scenarios():
    scenarios = [
        ("python", {"compressor": groupcompress.PythonGroupCompressor}),
    ]
    if _compiled_groupcompress_module is not None:
        scenarios.append(("C", {"compressor": groupcompress.PyrexGroupCompressor}))
    return scenarios


load_tests = load_tests_apply_scenarios


class TestGroupCompressor(TestCase):
    def _chunks_to_repr_lines(self, chunks):
        return "\n".join(map(repr, b"".join(chunks).split(b"\n")))

    def assertEqualDiffEncoded(self, expected, actual):
        """Compare the actual content to the expected content.

        :param expected: A group of chunks that we expect to see
        :param actual: The measured 'chunks'

        We will transform the chunks back into lines, and then run 'repr()'
        over them to handle non-ascii characters.
        """
        self.assertEqualDiff(
            self._chunks_to_repr_lines(expected), self._chunks_to_repr_lines(actual)
        )


class TestAllGroupCompressors(TestGroupCompressor):
    """Tests for GroupCompressor."""

    scenarios = group_compress_implementation_scenarios()
    compressor = None  # Set by scenario

    def test_empty_delta(self):
        compressor = self.compressor()
        self.assertEqual([], compressor.chunks)

    def test_one_nosha_delta(self):
        # diff against NUKK
        compressor = self.compressor()
        text = b"strange\ncommon\n"
        sha1, start_point, end_point, _ = compressor.compress(
            (b"label",), [text], len(text), None
        )
        self.assertEqual(sha_string(b"strange\ncommon\n"), sha1)
        expected_lines = b"f\x0fstrange\ncommon\n"
        self.assertEqual(expected_lines, b"".join(compressor.chunks))
        self.assertEqual(0, start_point)
        self.assertEqual(len(expected_lines), end_point)

    def test_empty_content(self):
        compressor = self.compressor()
        # Adding empty bytes should return the 'null' record
        sha1, start_point, end_point, kind = compressor.compress(
            (b"empty",), [], 0, None
        )
        self.assertEqual(0, start_point)
        self.assertEqual(0, end_point)
        self.assertEqual("fulltext", kind)
        self.assertEqual(groupcompress._null_sha1, sha1)
        self.assertEqual(0, compressor.endpoint)
        self.assertEqual([], compressor.chunks)
        # Even after adding some content
        text = b"some\nbytes\n"
        compressor.compress((b"content",), [text], len(text), None)
        self.assertGreater(compressor.endpoint, 0)
        sha1, start_point, end_point, kind = compressor.compress(
            (b"empty2",), [], 0, None
        )
        self.assertEqual(0, start_point)
        self.assertEqual(0, end_point)
        self.assertEqual("fulltext", kind)
        self.assertEqual(groupcompress._null_sha1, sha1)

    def test_extract_from_compressor(self):
        # Knit fetching will try to reconstruct texts locally which results in
        # reading something that is in the compressor stream already.
        compressor = self.compressor()
        text = b"strange\ncommon long line\nthat needs a 16 byte match\n"
        sha1_1, _, _, _ = compressor.compress((b"label",), [text], len(text), None)
        list(compressor.chunks)
        text = b"common long line\nthat needs a 16 byte match\ndifferent\n"
        sha1_2, _, _end_point, _ = compressor.compress(
            (b"newlabel",), [text], len(text), None
        )
        # get the first out
        self.assertEqual(
            ([b"strange\ncommon long line\nthat needs a 16 byte match\n"], sha1_1),
            compressor.extract((b"label",)),
        )
        # and the second
        self.assertEqual(
            (
                [b"common long line\nthat needs a 16 byte match\ndifferent\n"],
                sha1_2,
            ),
            compressor.extract((b"newlabel",)),
        )


class TestPyrexGroupCompressor(TestGroupCompressor):
    compressor = groupcompress.PyrexGroupCompressor

    def setUp(self):
        super().setUp()
        if _compiled_groupcompress_module is None:
            self.skipTest("bzrformats._groupcompress_pyx not available")

    def test_stats(self):
        compressor = self.compressor()
        chunks = [b"strange\n", b"common very very long line\n", b"plus more text\n"]
        compressor.compress((b"label",), chunks, sum(map(len, chunks)), None)
        chunks = [
            b"common very very long line\n",
            b"plus more text\n",
            b"different\n",
            b"moredifferent\n",
        ]
        compressor.compress((b"newlabel",), chunks, sum(map(len, chunks)), None)
        chunks = [
            b"new\n",
            b"common very very long line\n",
            b"plus more text\n",
            b"different\n",
            b"moredifferent\n",
        ]
        compressor.compress((b"label3",), chunks, sum(map(len, chunks)), None)
        self.assertAlmostEqual(1.9, compressor.ratio(), 1)

    def test_two_nosha_delta(self):
        compressor = self.compressor()
        text = b"strange\ncommon long line\nthat needs a 16 byte match\n"
        _sha1_1, _, _, _ = compressor.compress((b"label",), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = b"common long line\nthat needs a 16 byte match\ndifferent\n"
        sha1_2, _start_point, end_point, _ = compressor.compress(
            (b"newlabel",), [text], len(text), None
        )
        self.assertEqual(sha_string(text), sha1_2)
        expected_lines.extend(
            [
                # 'delta', delta length
                b"d\x0f",
                # source and target length
                b"\x36",
                # copy the line common
                b"\x91\x0a\x2c",  # copy, offset 0x0a, len 0x2c
                # add the line different, and the trailing newline
                b"\x0adifferent\n",  # insert 10 bytes
            ]
        )
        self.assertEqualDiffEncoded(expected_lines, compressor.chunks)
        self.assertEqual(sum(map(len, expected_lines)), end_point)

    def test_three_nosha_delta(self):
        # The first interesting test: make a change that should use lines from
        # both parents.
        compressor = self.compressor()
        text = b"strange\ncommon very very long line\nwith some extra text\n"
        _sha1_1, _, _, _ = compressor.compress((b"label",), [text], len(text), None)
        text = b"different\nmoredifferent\nand then some more\n"
        _sha1_2, _, _, _ = compressor.compress((b"newlabel",), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = (
            b"new\ncommon very very long line\nwith some extra text\n"
            b"different\nmoredifferent\nand then some more\n"
        )
        sha1_3, _start_point, end_point, _ = compressor.compress(
            (b"label3",), [text], len(text), None
        )
        self.assertEqual(sha_string(text), sha1_3)
        expected_lines.extend(
            [
                # 'delta', delta length
                b"d\x0b",
                # source and target length
                b"\x5f"
                # insert new
                b"\x03new",
                # Copy of first parent 'common' range
                b"\x91\x09\x31"  # copy, offset 0x09, 0x31 bytes
                # Copy of second parent 'different' range
                b"\x91\x3c\x2b",  # copy, offset 0x3c, 0x2b bytes
            ]
        )
        self.assertEqualDiffEncoded(expected_lines, compressor.chunks)
        self.assertEqual(sum(map(len, expected_lines)), end_point)


class TestPythonGroupCompressor(TestGroupCompressor):
    compressor = groupcompress.PythonGroupCompressor

    def test_stats(self):
        compressor = self.compressor()
        chunks = [b"strange\n", b"common very very long line\n", b"plus more text\n"]
        compressor.compress((b"label",), chunks, sum(map(len, chunks)), None)
        chunks = [
            b"common very very long line\n",
            b"plus more text\n",
            b"different\n",
            b"moredifferent\n",
        ]
        compressor.compress((b"newlabel",), chunks, sum(map(len, chunks)), None)
        chunks = [
            b"new\n",
            b"common very very long line\n",
            b"plus more text\n",
            b"different\n",
            b"moredifferent\n",
        ]
        compressor.compress((b"label3",), chunks, sum(map(len, chunks)), None)
        self.assertAlmostEqual(1.9, compressor.ratio(), 1)

    def test_two_nosha_delta(self):
        compressor = self.compressor()
        text = b"strange\ncommon long line\nthat needs a 16 byte match\n"
        _sha1_1, _, _, _ = compressor.compress((b"label",), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = b"common long line\nthat needs a 16 byte match\ndifferent\n"
        sha1_2, _start_point, end_point, _ = compressor.compress(
            (b"newlabel",), [text], len(text), None
        )
        self.assertEqual(sha_string(text), sha1_2)
        expected_lines.extend(
            [
                # 'delta', delta length
                b"d\x0f",
                # target length
                b"\x36",
                # copy the line common
                b"\x91\x0a\x2c",  # copy, offset 0x0a, len 0x2c
                # add the line different, and the trailing newline
                b"\x0adifferent\n",  # insert 10 bytes
            ]
        )
        self.assertEqualDiffEncoded(expected_lines, compressor.chunks)
        self.assertEqual(sum(map(len, expected_lines)), end_point)

    def test_three_nosha_delta(self):
        # The first interesting test: make a change that should use lines from
        # both parents.
        compressor = self.compressor()
        text = b"strange\ncommon very very long line\nwith some extra text\n"
        _sha1_1, _, _, _ = compressor.compress((b"label",), [text], len(text), None)
        text = b"different\nmoredifferent\nand then some more\n"
        _sha1_2, _, _, _ = compressor.compress((b"newlabel",), [text], len(text), None)
        expected_lines = list(compressor.chunks)
        text = (
            b"new\ncommon very very long line\nwith some extra text\n"
            b"different\nmoredifferent\nand then some more\n"
        )
        sha1_3, _start_point, end_point, _ = compressor.compress(
            (b"label3",), [text], len(text), None
        )
        self.assertEqual(sha_string(text), sha1_3)
        expected_lines.extend(
            [
                # 'delta', delta length
                b"d\x0c",
                # target length
                b"\x5f"
                # insert new
                b"\x04new\n",
                # Copy of first parent 'common' range
                b"\x91\x0a\x30"  # copy, offset 0x0a, 0x30 bytes
                # Copy of second parent 'different' range
                b"\x91\x3c\x2b",  # copy, offset 0x3c, 0x2b bytes
            ]
        )
        self.assertEqualDiffEncoded(expected_lines, compressor.chunks)
        self.assertEqual(sum(map(len, expected_lines)), end_point)


class TestGroupCompressBlock(TestCase):
    def make_block(self, key_to_text):
        """Create a GroupCompressBlock, filling it with the given texts."""
        compressor = groupcompress.GroupCompressor()
        for key in sorted(key_to_text):
            compressor.compress(key, [key_to_text[key]], len(key_to_text[key]), None)
        locs = {
            key: (start, end)
            for key, (start, _, end, _) in compressor.labels_deltas.items()
        }
        block = compressor.flush()
        raw_bytes = block.to_bytes()
        # Go through from_bytes(to_bytes()) so that we start with a compressed
        # content object
        return locs, groupcompress.GroupCompressBlock.from_bytes(raw_bytes)

    def test_from_empty_bytes(self):
        self.assertRaises(ValueError, groupcompress.GroupCompressBlock.from_bytes, b"")

    def test_from_minimal_bytes(self):
        block = groupcompress.GroupCompressBlock.from_bytes(b"gcb1z\n0\n0\n")
        self.assertIsInstance(block, groupcompress.GroupCompressBlock)
        self.assertIs(None, block._content)
        self.assertEqual(b"", block._z_content)
        block._ensure_content()
        self.assertEqual(b"", block._content)
        self.assertEqual(b"", block._z_content)
        block._ensure_content()  # Ensure content is safe to call 2x

    def test_from_invalid(self):
        self.assertRaises(
            ValueError,
            groupcompress.GroupCompressBlock.from_bytes,
            b"this is not a valid header",
        )

    def test_from_bytes(self):
        content = b"a tiny bit of content\n"
        z_content = zlib.compress(content)
        z_bytes = (
            b"gcb1z\n"  # group compress block v1 plain
            b"%d\n"  # Length of compressed content
            b"%d\n"  # Length of uncompressed content
            b"%s"  # Compressed content
        ) % (len(z_content), len(content), z_content)
        block = groupcompress.GroupCompressBlock.from_bytes(z_bytes)
        self.assertEqual(z_content, block._z_content)
        self.assertIs(None, block._content)
        self.assertEqual(len(z_content), block._z_content_length)
        self.assertEqual(len(content), block._content_length)
        block._ensure_content()
        self.assertEqual(z_content, block._z_content)
        self.assertEqual(content, block._content)

    def test_to_chunks(self):
        content_chunks = [
            b"this is some content\n",
            b"this content will be compressed\n",
        ]
        content_len = sum(map(len, content_chunks))
        content = b"".join(content_chunks)
        gcb = groupcompress.GroupCompressBlock()
        gcb.set_chunked_content(content_chunks, content_len)
        total_len, block_chunks = gcb.to_chunks()
        block_bytes = b"".join(block_chunks)
        self.assertEqual(gcb._z_content_length, len(gcb._z_content))
        self.assertEqual(total_len, len(block_bytes))
        self.assertEqual(gcb._content_length, content_len)
        expected_header = (
            b"gcb1z\n"  # group compress block v1 zlib
            b"%d\n"  # Length of compressed content
            b"%d\n"  # Length of uncompressed content
        ) % (gcb._z_content_length, gcb._content_length)
        # The first chunk should be the header chunk. It is small, fixed size,
        # and there is no compelling reason to split it up
        self.assertEqual(expected_header, block_chunks[0])
        self.assertStartsWith(block_bytes, expected_header)
        remaining_bytes = block_bytes[len(expected_header) :]
        raw_bytes = zlib.decompress(remaining_bytes)
        self.assertEqual(content, raw_bytes)

    def test_to_bytes(self):
        content = b"this is some content\nthis content will be compressed\n"
        gcb = groupcompress.GroupCompressBlock()
        gcb.set_content(content)
        data = gcb.to_bytes()
        self.assertEqual(gcb._z_content_length, len(gcb._z_content))
        self.assertEqual(gcb._content_length, len(content))
        expected_header = (
            b"gcb1z\n"  # group compress block v1 zlib
            b"%d\n"  # Length of compressed content
            b"%d\n"  # Length of uncompressed content
        ) % (gcb._z_content_length, gcb._content_length)
        self.assertStartsWith(data, expected_header)
        remaining_bytes = data[len(expected_header) :]
        raw_bytes = zlib.decompress(remaining_bytes)
        self.assertEqual(content, raw_bytes)

        # we should get the same results if using the chunked version
        gcb = groupcompress.GroupCompressBlock()
        gcb.set_chunked_content(
            [b"this is some content\nthis content will be compressed\n"],
            len(content),
        )
        old_data = data
        data = gcb.to_bytes()
        self.assertEqual(old_data, data)

    def test_partial_decomp(self):
        content_chunks = []
        # We need a sufficient amount of data so that zlib.decompress has
        # partial decompression to work with. Most auto-generated data
        # compresses a bit too well, we want a combination, so we combine a sha
        # hash with compressible data.
        for i in range(2048):
            next_content = b"%d\nThis is a bit of duplicate text\n" % (i,)
            content_chunks.append(next_content)
            next_sha1 = osutils.sha_string(next_content)
            content_chunks.append(next_sha1 + b"\n")
        content = b"".join(content_chunks)
        self.assertEqual(158634, len(content))
        z_content = zlib.compress(content)
        self.assertEqual(57182, len(z_content))
        block = groupcompress.GroupCompressBlock()
        block._z_content_chunks = (z_content,)
        block._z_content_length = len(z_content)
        block._compressor_name = "zlib"
        block._content_length = 158634
        self.assertIs(None, block._content)
        block._ensure_content(100)
        self.assertIsNot(None, block._content)
        # We have decompressed at least 100 bytes
        self.assertGreaterEqual(len(block._content), 100)
        # We have not decompressed the whole content
        self.assertLess(len(block._content), 158634)
        self.assertEqualDiff(content[: len(block._content)], block._content)
        # ensuring content that we already have shouldn't cause any more data
        # to be extracted
        cur_len = len(block._content)
        block._ensure_content(cur_len - 10)
        self.assertEqual(cur_len, len(block._content))
        # Now we want a bit more content
        cur_len += 10
        block._ensure_content(cur_len)
        self.assertGreaterEqual(len(block._content), cur_len)
        self.assertLess(len(block._content), 158634)
        self.assertEqualDiff(content[: len(block._content)], block._content)
        # And now lets finish
        block._ensure_content(158634)
        self.assertEqualDiff(content, block._content)
        # And the decompressor is finalized
        self.assertIs(None, block._z_content_decompressor)

    def test__ensure_all_content(self):
        content_chunks = []
        # We need a sufficient amount of data so that zlib.decompress has
        # partial decompression to work with. Most auto-generated data
        # compresses a bit too well, we want a combination, so we combine a sha
        # hash with compressible data.
        for i in range(2048):
            next_content = b"%d\nThis is a bit of duplicate text\n" % (i,)
            content_chunks.append(next_content)
            next_sha1 = osutils.sha_string(next_content)
            content_chunks.append(next_sha1 + b"\n")
        content = b"".join(content_chunks)
        self.assertEqual(158634, len(content))
        z_content = zlib.compress(content)
        self.assertEqual(57182, len(z_content))
        block = groupcompress.GroupCompressBlock()
        block._z_content_chunks = (z_content,)
        block._z_content_length = len(z_content)
        block._compressor_name = "zlib"
        block._content_length = 158634
        self.assertIs(None, block._content)
        # The first _ensure_content got all of the required data
        block._ensure_content(158634)
        self.assertEqualDiff(content, block._content)
        # And we should have released the _z_content_decompressor since it was
        # fully consumed
        self.assertIs(None, block._z_content_decompressor)

    def test__dump(self):
        dup_content = b"some duplicate content\nwhich is sufficiently long\n"
        key_to_text = {
            (b"1",): dup_content + b"1 unique\n",
            (b"2",): dup_content + b"2 extra special\n",
        }
        _locs, block = self.make_block(key_to_text)
        self.assertEqual(
            [
                (b"f", len(key_to_text[(b"1",)])),
                (
                    b"d",
                    21,
                    len(key_to_text[(b"2",)]),
                    [
                        (b"c", 2, len(dup_content)),
                        (b"i", len(b"2 extra special\n"), b""),
                    ],
                ),
            ],
            block._dump(),
        )


class TestCaseWithGroupCompressVersionedFiles(TestCaseWithMemoryTransport):
    def make_test_vf(
        self,
        create_graph,
        keylength=1,
        do_cleanup=True,
        dir=".",
        inconsistency_fatal=True,
    ):
        t = self.get_transport(dir)
        t.ensure_base()
        vf = groupcompress.make_pack_factory(
            graph=create_graph,
            delta=False,
            keylength=keylength,
            inconsistency_fatal=inconsistency_fatal,
        )(t)
        if do_cleanup:
            self.addCleanup(groupcompress.cleanup_pack_group, vf)
        return vf


class TestGroupCompressVersionedFiles(TestCaseWithGroupCompressVersionedFiles):
    def make_g_index(self, name, ref_lists=0, nodes=None):
        if nodes is None:
            nodes = []
        builder = btree_index.BTreeBuilder(ref_lists)
        for node, references, value in nodes:
            builder.add_node(node, references, value)
        stream = builder.finish()
        trans = self.get_transport()
        size = trans.put_file(name, stream)
        return btree_index.BTreeGraphIndex(trans, name, size)

    def make_g_index_missing_parent(self):
        graph_index = self.make_g_index(
            "missing_parent",
            1,
            [
                ((b"parent",), b"2 78 2 10", ([],)),
                ((b"tip",), b"2 78 2 10", ([(b"parent",), (b"missing-parent",)],)),
            ],
        )
        return graph_index

    def test_get_record_stream_as_requested(self):
        # Consider promoting 'as-requested' to general availability, and
        # make this a VF interface test
        vf = self.make_test_vf(False, dir="source")
        vf.add_lines((b"a",), (), [b"lines\n"])
        vf.add_lines((b"b",), (), [b"lines\n"])
        vf.add_lines((b"c",), (), [b"lines\n"])
        vf.add_lines((b"d",), (), [b"lines\n"])
        vf.writer.end()
        keys = [
            record.key
            for record in vf.get_record_stream(
                [(b"a",), (b"b",), (b"c",), (b"d",)], "as-requested", False
            )
        ]
        self.assertEqual([(b"a",), (b"b",), (b"c",), (b"d",)], keys)
        keys = [
            record.key
            for record in vf.get_record_stream(
                [(b"b",), (b"a",), (b"d",), (b"c",)], "as-requested", False
            )
        ]
        self.assertEqual([(b"b",), (b"a",), (b"d",), (b"c",)], keys)

        # It should work even after being repacked into another VF
        vf2 = self.make_test_vf(False, dir="target")
        vf2.insert_record_stream(
            vf.get_record_stream(
                [(b"b",), (b"a",), (b"d",), (b"c",)], "as-requested", False
            )
        )
        vf2.writer.end()

        keys = [
            record.key
            for record in vf2.get_record_stream(
                [(b"a",), (b"b",), (b"c",), (b"d",)], "as-requested", False
            )
        ]
        self.assertEqual([(b"a",), (b"b",), (b"c",), (b"d",)], keys)
        keys = [
            record.key
            for record in vf2.get_record_stream(
                [(b"b",), (b"a",), (b"d",), (b"c",)], "as-requested", False
            )
        ]
        self.assertEqual([(b"b",), (b"a",), (b"d",), (b"c",)], keys)

    def test_get_record_stream_max_bytes_to_index_default(self):
        vf = self.make_test_vf(True, dir="source")
        vf.add_lines((b"a",), (), [b"lines\n"])
        vf.writer.end()
        record = next(vf.get_record_stream([(b"a",)], "unordered", True))
        self.assertEqual(
            vf._DEFAULT_COMPRESSOR_SETTINGS, record._manager._get_compressor_settings()
        )

    def test_get_record_stream_accesses_compressor_settings(self):
        vf = self.make_test_vf(True, dir="source")
        vf.add_lines((b"a",), (), [b"lines\n"])
        vf.writer.end()
        vf._max_bytes_to_index = 1234
        record = next(vf.get_record_stream([(b"a",)], "unordered", True))
        self.assertEqual(
            {"max_bytes_to_index": 1234}, record._manager._get_compressor_settings()
        )

    @staticmethod
    def grouped_stream(revision_ids, first_parents=()):
        parents = first_parents
        for revision_id in revision_ids:
            key = (revision_id,)
            record = versionedfile.FulltextContentFactory(
                key,
                parents,
                None,
                b"some content that is\n"
                b"identical except for\n"
                b"revision_id:%s\n" % (revision_id,),
            )
            yield record
            parents = (key,)

    def test_insert_record_stream_reuses_blocks(self):
        vf = self.make_test_vf(True, dir="source")
        # One group, a-d
        vf.insert_record_stream(self.grouped_stream([b"a", b"b", b"c", b"d"]))
        # Second group, e-h
        vf.insert_record_stream(
            self.grouped_stream([b"e", b"f", b"g", b"h"], first_parents=((b"d",),))
        )
        block_bytes = {}
        stream = vf.get_record_stream(
            [(r.encode(),) for r in "abcdefgh"], "unordered", False
        )
        num_records = 0
        for record in stream:
            if record.key in [(b"a",), (b"e",)]:
                self.assertEqual("groupcompress-block", record.storage_kind)
            else:
                self.assertEqual("groupcompress-block-ref", record.storage_kind)
            block_bytes[record.key] = record._manager._block._z_content
            num_records += 1
        self.assertEqual(8, num_records)
        for r in "abcd":
            key = (r.encode(),)
            self.assertIs(block_bytes[key], block_bytes[(b"a",)])
            self.assertNotEqual(block_bytes[key], block_bytes[(b"e",)])
        for r in "efgh":
            key = (r.encode(),)
            self.assertIs(block_bytes[key], block_bytes[(b"e",)])
            self.assertNotEqual(block_bytes[key], block_bytes[(b"a",)])
        # Now copy the blocks into another vf, and ensure that the blocks are
        # preserved without creating new entries
        vf2 = self.make_test_vf(True, dir="target")
        keys = [(r.encode(),) for r in "abcdefgh"]
        # ordering in 'groupcompress' order, should actually swap the groups in
        # the target vf, but the groups themselves should not be disturbed.

        def small_size_stream():
            for record in vf.get_record_stream(keys, "groupcompress", False):
                record._manager._full_enough_block_size = (
                    record._manager._block._content_length
                )
                yield record

        vf2.insert_record_stream(small_size_stream())
        stream = vf2.get_record_stream(keys, "groupcompress", False)
        vf2.writer.end()
        num_records = 0
        for record in stream:
            num_records += 1
            self.assertEqual(block_bytes[record.key], record._manager._block._z_content)
        self.assertEqual(8, num_records)

    def test_insert_record_stream_packs_on_the_fly(self):
        vf = self.make_test_vf(True, dir="source")
        # One group, a-d
        vf.insert_record_stream(self.grouped_stream([b"a", b"b", b"c", b"d"]))
        # Second group, e-h
        vf.insert_record_stream(
            self.grouped_stream([b"e", b"f", b"g", b"h"], first_parents=((b"d",),))
        )
        # Now copy the blocks into another vf, and see that the
        # insert_record_stream rebuilt a new block on-the-fly because of
        # under-utilization
        vf2 = self.make_test_vf(True, dir="target")
        keys = [(r.encode(),) for r in "abcdefgh"]
        vf2.insert_record_stream(vf.get_record_stream(keys, "groupcompress", False))
        stream = vf2.get_record_stream(keys, "groupcompress", False)
        vf2.writer.end()
        num_records = 0
        # All of the records should be recombined into a single block
        block = None
        for record in stream:
            num_records += 1
            if block is None:
                block = record._manager._block
            else:
                self.assertIs(block, record._manager._block)
        self.assertEqual(8, num_records)

    def test__insert_record_stream_no_reuse_block(self):
        vf = self.make_test_vf(True, dir="source")
        # One group, a-d
        vf.insert_record_stream(self.grouped_stream([b"a", b"b", b"c", b"d"]))
        # Second group, e-h
        vf.insert_record_stream(
            self.grouped_stream([b"e", b"f", b"g", b"h"], first_parents=((b"d",),))
        )
        vf.writer.end()
        keys = [(r.encode(),) for r in "abcdefgh"]
        self.assertEqual(8, len(list(vf.get_record_stream(keys, "unordered", False))))
        # Now copy the blocks into another vf, and ensure that the blocks are
        # preserved without creating new entries
        vf2 = self.make_test_vf(True, dir="target")
        # ordering in 'groupcompress' order, should actually swap the groups in
        # the target vf, but the groups themselves should not be disturbed.
        list(
            vf2._insert_record_stream(
                vf.get_record_stream(keys, "groupcompress", False), reuse_blocks=False
            )
        )
        vf2.writer.end()
        # After inserting with reuse_blocks=False, we should have everything in
        # a single new block.
        stream = vf2.get_record_stream(keys, "groupcompress", False)
        block = None
        for record in stream:
            if block is None:
                block = record._manager._block
            else:
                self.assertIs(block, record._manager._block)

    def test_add_missing_noncompression_parent_unvalidated_index(self):
        unvalidated = self.make_g_index_missing_parent()
        combined = _mod_index.CombinedGraphIndex([unvalidated])
        index = groupcompress._GCGraphIndex(
            combined,
            is_locked=lambda: True,
            parents=True,
            track_external_parent_refs=True,
        )
        index.scan_unvalidated_index(unvalidated)
        self.assertEqual(frozenset([(b"missing-parent",)]), index.get_missing_parents())

    def test_track_external_parent_refs(self):
        g_index = self.make_g_index("empty", 1, [])
        mod_index = btree_index.BTreeBuilder(1, 1)
        combined = _mod_index.CombinedGraphIndex([g_index, mod_index])
        index = groupcompress._GCGraphIndex(
            combined,
            is_locked=lambda: True,
            parents=True,
            add_callback=mod_index.add_nodes,
            track_external_parent_refs=True,
        )
        index.add_records(
            [((b"new-key",), b"2 10 2 10", [((b"parent-1",), (b"parent-2",))])]
        )
        self.assertEqual(
            frozenset([(b"parent-1",), (b"parent-2",)]), index.get_missing_parents()
        )

    def make_source_with_b(self, a_parent, path):
        source = self.make_test_vf(True, dir=path)
        source.add_lines((b"a",), (), [b"lines\n"])
        b_parents = ((b"a",),) if a_parent else ()
        source.add_lines((b"b",), b_parents, [b"lines\n"])
        return source

    def do_inconsistent_inserts(self, inconsistency_fatal):
        target = self.make_test_vf(
            True, dir="target", inconsistency_fatal=inconsistency_fatal
        )
        for x in range(2):
            source = self.make_source_with_b(x == 1, f"source{x}")
            target.insert_record_stream(
                source.get_record_stream([(b"b",)], "unordered", False)
            )

    def test_inconsistent_redundant_inserts_warn(self):
        """Should not insert a record that is already present."""
        # Capture logging warnings
        import io

        log_stream = io.StringIO()
        handler = logging.StreamHandler(log_stream)
        handler.setLevel(logging.WARNING)

        # Get the groupcompress logger
        gc_logger = logging.getLogger("bzrformats.groupcompress")
        gc_logger.addHandler(handler)
        old_level = gc_logger.level
        gc_logger.setLevel(logging.WARNING)

        try:
            self.do_inconsistent_inserts(inconsistency_fatal=False)
        finally:
            gc_logger.removeHandler(handler)
            gc_logger.setLevel(old_level)

        warnings = log_stream.getvalue()
        self.assertContainsRe(
            warnings,
            r"inconsistent details in skipped record: \(b?'b',\)"
            r" \(b?'42 32 0 8', \(\(\),\)\)"
            r" \(b?'74 32 0 8', \(\(\(b?'a',\),\),\)\)$",
        )

    def test_inconsistent_redundant_inserts_raises(self):
        e = self.assertRaises(
            knit.KnitCorrupt, self.do_inconsistent_inserts, inconsistency_fatal=True
        )
        self.assertContainsRe(
            str(e),
            r"Knit.* corrupt: inconsistent details"
            r" in add_records:"
            r" \(b?'b',\) \(b?'42 32 0 8', \(\(\),\)\)"
            r" \(b?'74 32 0 8', \(\(\(b?'a',\),\),\)\)",
        )

    def test_clear_cache(self):
        vf = self.make_source_with_b(True, "source")
        vf.writer.end()
        for _record in vf.get_record_stream([(b"a",), (b"b",)], "unordered", True):
            pass
        self.assertGreater(len(vf._group_cache), 0)
        vf.clear_cache()
        self.assertEqual(0, len(vf._group_cache))


class TestGroupCompressConfig(TestCaseWithMemoryTransport):
    def make_test_vf(self):
        t = self.get_transport(".")
        t.ensure_base()
        factory = groupcompress.make_pack_factory(
            graph=True, delta=False, keylength=1, inconsistency_fatal=True
        )
        vf = factory(t)
        self.addCleanup(groupcompress.cleanup_pack_group, vf)
        return vf

    def test_max_bytes_to_index_default(self):
        vf = self.make_test_vf()
        gc = vf._make_group_compressor()
        self.assertEqual(vf._DEFAULT_MAX_BYTES_TO_INDEX, vf._max_bytes_to_index)
        if isinstance(gc, groupcompress.PyrexGroupCompressor):
            self.assertEqual(
                vf._DEFAULT_MAX_BYTES_TO_INDEX, gc._delta_index._max_bytes_to_index
            )

    def test_max_bytes_to_index_set_directly(self):
        vf = self.make_test_vf()
        vf._max_bytes_to_index = 10000
        gc = vf._make_group_compressor()
        self.assertEqual(10000, vf._max_bytes_to_index)
        if isinstance(gc, groupcompress.PyrexGroupCompressor):
            self.assertEqual(10000, gc._delta_index._max_bytes_to_index)


class StubGCVF:
    def __init__(self, canned_get_blocks=None):
        self._group_cache = {}
        self._canned_get_blocks = canned_get_blocks or []

    def _get_blocks(self, read_memos):
        return iter(self._canned_get_blocks)


class Test_BatchingBlockFetcher(TestCaseWithGroupCompressVersionedFiles):
    """Simple whitebox unit tests for _BatchingBlockFetcher."""

    def test_add_key_new_read_memo(self):
        """Adding a key with an uncached read_memo new to this batch adds that
        read_memo to the list of memos to fetch.
        """
        # locations are: index_memo, ignored, parents, ignored
        # where index_memo is: (idx, offset, len, factory_start, factory_end)
        # and (idx, offset, size) is known as the 'read_memo', identifying the
        # raw bytes needed.
        read_memo = ("fake index", 100, 50)
        locations = {("key",): (read_memo + (None, None), None, None, None)}
        batcher = groupcompress._BatchingBlockFetcher(StubGCVF(), locations)
        total_size = batcher.add_key(("key",))
        self.assertEqual(50, total_size)
        self.assertEqual([("key",)], batcher.keys)
        self.assertEqual([read_memo], batcher.memos_to_get)

    def test_add_key_duplicate_read_memo(self):
        """read_memos that occur multiple times in a batch will only be fetched
        once.
        """
        read_memo = ("fake index", 100, 50)
        # Two keys, both sharing the same read memo (but different overall
        # index_memos).
        locations = {
            ("key1",): (read_memo + (0, 1), None, None, None),
            ("key2",): (read_memo + (1, 2), None, None, None),
        }
        batcher = groupcompress._BatchingBlockFetcher(StubGCVF(), locations)
        total_size = batcher.add_key(("key1",))
        total_size = batcher.add_key(("key2",))
        self.assertEqual(50, total_size)
        self.assertEqual([("key1",), ("key2",)], batcher.keys)
        self.assertEqual([read_memo], batcher.memos_to_get)

    def test_add_key_cached_read_memo(self):
        """Adding a key with a cached read_memo will not cause that read_memo
        to be added to the list to fetch.
        """
        read_memo = ("fake index", 100, 50)
        gcvf = StubGCVF()
        gcvf._group_cache[read_memo] = "fake block"
        locations = {("key",): (read_memo + (None, None), None, None, None)}
        batcher = groupcompress._BatchingBlockFetcher(gcvf, locations)
        total_size = batcher.add_key(("key",))
        self.assertEqual(0, total_size)
        self.assertEqual([("key",)], batcher.keys)
        self.assertEqual([], batcher.memos_to_get)

    def test_yield_factories_empty(self):
        """An empty batch yields no factories."""
        batcher = groupcompress._BatchingBlockFetcher(StubGCVF(), {})
        self.assertEqual([], list(batcher.yield_factories()))

    def test_yield_factories_calls_get_blocks(self):
        """Uncached memos are retrieved via get_blocks."""
        read_memo1 = ("fake index", 100, 50)
        read_memo2 = ("fake index", 150, 40)
        gcvf = StubGCVF(
            canned_get_blocks=[
                (read_memo1, groupcompress.GroupCompressBlock()),
                (read_memo2, groupcompress.GroupCompressBlock()),
            ]
        )
        locations = {
            ("key1",): (read_memo1 + (0, 0), None, None, None),
            ("key2",): (read_memo2 + (0, 0), None, None, None),
        }
        batcher = groupcompress._BatchingBlockFetcher(gcvf, locations)
        batcher.add_key(("key1",))
        batcher.add_key(("key2",))
        factories = list(batcher.yield_factories(full_flush=True))
        self.assertLength(2, factories)
        keys = [f.key for f in factories]
        kinds = [f.storage_kind for f in factories]
        self.assertEqual([("key1",), ("key2",)], keys)
        self.assertEqual(["groupcompress-block", "groupcompress-block"], kinds)

    def test_yield_factories_flushing(self):
        """yield_factories holds back on yielding results from the final block
        unless passed full_flush=True.
        """
        fake_block = groupcompress.GroupCompressBlock()
        read_memo = ("fake index", 100, 50)
        gcvf = StubGCVF()
        gcvf._group_cache[read_memo] = fake_block
        locations = {("key",): (read_memo + (0, 0), None, None, None)}
        batcher = groupcompress._BatchingBlockFetcher(gcvf, locations)
        batcher.add_key(("key",))
        self.assertEqual([], list(batcher.yield_factories()))
        factories = list(batcher.yield_factories(full_flush=True))
        self.assertLength(1, factories)
        self.assertEqual(("key",), factories[0].key)
        self.assertEqual("groupcompress-block", factories[0].storage_kind)


class TestLazyGroupCompress(TestCaseWithMemoryTransport):
    _texts = {
        (b"key1",): b"this is a text\n"
        b"with a reasonable amount of compressible bytes\n"
        b"which can be shared between various other texts\n",
        (b"key2",): b"another text\n"
        b"with a reasonable amount of compressible bytes\n"
        b"which can be shared between various other texts\n",
        (b"key3",): b"yet another text which won't be extracted\n"
        b"with a reasonable amount of compressible bytes\n"
        b"which can be shared between various other texts\n",
        (b"key4",): b"this will be extracted\n"
        b"but references most of its bytes from\n"
        b"yet another text which won't be extracted\n"
        b"with a reasonable amount of compressible bytes\n"
        b"which can be shared between various other texts\n",
    }

    def make_block(self, key_to_text):
        """Create a GroupCompressBlock, filling it with the given texts."""
        compressor = groupcompress.GroupCompressor()
        for key in sorted(key_to_text):
            compressor.compress(key, [key_to_text[key]], len(key_to_text[key]), None)
        locs = {
            key: (start, end)
            for key, (start, _, end, _) in compressor.labels_deltas.items()
        }
        block = compressor.flush()
        raw_bytes = block.to_bytes()
        return locs, groupcompress.GroupCompressBlock.from_bytes(raw_bytes)

    def add_key_to_manager(self, key, locations, block, manager):
        start, end = locations[key]
        manager.add_factory(key, (), start, end)

    def make_block_and_full_manager(self, texts):
        locations, block = self.make_block(texts)
        manager = groupcompress._LazyGroupContentManager(block)
        for key in sorted(texts):
            self.add_key_to_manager(key, locations, block, manager)
        return block, manager

    def test_get_fulltexts(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        self.add_key_to_manager((b"key1",), locations, block, manager)
        self.add_key_to_manager((b"key2",), locations, block, manager)
        result_order = []
        for record in manager.get_record_stream():
            result_order.append(record.key)
            text = self._texts[record.key]
            self.assertEqual(text, record.get_bytes_as("fulltext"))
        self.assertEqual([(b"key1",), (b"key2",)], result_order)

        # If we build the manager in the opposite order, we should get them
        # back in the opposite order
        manager = groupcompress._LazyGroupContentManager(block)
        self.add_key_to_manager((b"key2",), locations, block, manager)
        self.add_key_to_manager((b"key1",), locations, block, manager)
        result_order = []
        for record in manager.get_record_stream():
            result_order.append(record.key)
            text = self._texts[record.key]
            self.assertEqual(text, record.get_bytes_as("fulltext"))
        self.assertEqual([(b"key2",), (b"key1",)], result_order)

    def test__wire_bytes_no_keys(self):
        _locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        wire_bytes = manager._wire_bytes()
        block_length = len(block.to_bytes())
        # We should have triggered a strip, since we aren't using any content
        stripped_block = manager._block.to_bytes()
        self.assertGreater(block_length, len(stripped_block))
        empty_z_header = zlib.compress(b"")
        self.assertEqual(
            b"groupcompress-block\n"
            b"8\n"  # len(compress(''))
            b"0\n"  # len('')
            b"%d\n"  # compressed block len
            b"%s"  # zheader
            b"%s" % (len(stripped_block), empty_z_header, stripped_block),  # block
            wire_bytes,
        )

    def test__wire_bytes(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        self.add_key_to_manager((b"key1",), locations, block, manager)
        self.add_key_to_manager((b"key4",), locations, block, manager)
        block_bytes = block.to_bytes()
        wire_bytes = manager._wire_bytes()
        (storage_kind, z_header_len, header_len, block_len, rest) = wire_bytes.split(
            b"\n", 4
        )
        z_header_len = int(z_header_len)
        header_len = int(header_len)
        block_len = int(block_len)
        self.assertEqual(b"groupcompress-block", storage_kind)
        self.assertEqual(34, z_header_len)
        self.assertEqual(26, header_len)
        self.assertEqual(len(block_bytes), block_len)
        z_header = rest[:z_header_len]
        header = zlib.decompress(z_header)
        self.assertEqual(header_len, len(header))
        entry1 = locations[(b"key1",)]
        entry4 = locations[(b"key4",)]
        self.assertEqualDiff(
            b"key1\n"
            b"\n"  # no parents
            b"%d\n"  # start offset
            b"%d\n"  # end offset
            b"key4\n"
            b"\n"
            b"%d\n"
            b"%d\n" % (entry1[0], entry1[1], entry4[0], entry4[1]),
            header,
        )
        z_block = rest[z_header_len:]
        self.assertEqual(block_bytes, z_block)

    def test_from_bytes(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        self.add_key_to_manager((b"key1",), locations, block, manager)
        self.add_key_to_manager((b"key4",), locations, block, manager)
        wire_bytes = manager._wire_bytes()
        self.assertStartsWith(wire_bytes, b"groupcompress-block\n")
        manager = groupcompress._LazyGroupContentManager.from_bytes(wire_bytes)
        self.assertIsInstance(manager, groupcompress._LazyGroupContentManager)
        self.assertEqual(2, len(manager._factories))
        self.assertEqual(block._z_content, manager._block._z_content)
        result_order = []
        for record in manager.get_record_stream():
            result_order.append(record.key)
            text = self._texts[record.key]
            self.assertEqual(text, record.get_bytes_as("fulltext"))
        self.assertEqual([(b"key1",), (b"key4",)], result_order)

    def test__check_rebuild_no_changes(self):
        block, manager = self.make_block_and_full_manager(self._texts)
        manager._check_rebuild_block()
        self.assertIs(block, manager._block)

    def test__check_rebuild_only_one(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        # Request just the first key, which should trigger a 'strip' action
        self.add_key_to_manager((b"key1",), locations, block, manager)
        manager._check_rebuild_block()
        self.assertIsNot(block, manager._block)
        self.assertGreater(block._content_length, manager._block._content_length)
        # We should be able to still get the content out of this block, though
        # it should only have 1 entry
        for record in manager.get_record_stream():
            self.assertEqual((b"key1",), record.key)
            self.assertEqual(self._texts[record.key], record.get_bytes_as("fulltext"))

    def test__check_rebuild_middle(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        # Request a small key in the middle should trigger a 'rebuild'
        self.add_key_to_manager((b"key4",), locations, block, manager)
        manager._check_rebuild_block()
        self.assertIsNot(block, manager._block)
        self.assertGreater(block._content_length, manager._block._content_length)
        for record in manager.get_record_stream():
            self.assertEqual((b"key4",), record.key)
            self.assertEqual(self._texts[record.key], record.get_bytes_as("fulltext"))

    def test_manager_default_compressor_settings(self):
        _locations, old_block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(old_block)
        gcvf = groupcompress.GroupCompressVersionedFiles
        # It doesn't greedily evaluate _max_bytes_to_index
        self.assertIs(None, manager._compressor_settings)
        self.assertEqual(
            gcvf._DEFAULT_COMPRESSOR_SETTINGS, manager._get_compressor_settings()
        )

    def test_manager_custom_compressor_settings(self):
        _locations, old_block = self.make_block(self._texts)
        called = []

        def compressor_settings():
            called.append("called")
            return (10,)

        manager = groupcompress._LazyGroupContentManager(
            old_block, get_compressor_settings=compressor_settings
        )
        # It doesn't greedily evaluate compressor_settings
        self.assertIs(None, manager._compressor_settings)
        self.assertEqual((10,), manager._get_compressor_settings())
        self.assertEqual((10,), manager._get_compressor_settings())
        self.assertEqual((10,), manager._compressor_settings)
        # Only called 1 time
        self.assertEqual(["called"], called)

    def test__rebuild_handles_compressor_settings(self):
        if not isinstance(
            groupcompress.GroupCompressor, groupcompress.PyrexGroupCompressor
        ):
            raise TestNotApplicable(
                "pure-python compressor does not handle compressor_settings"
            )
        locations, old_block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(
            old_block, get_compressor_settings=lambda: {"max_bytes_to_index": 32}
        )
        gc = manager._make_group_compressor()
        self.assertEqual(32, gc._delta_index._max_bytes_to_index)
        self.add_key_to_manager((b"key3",), locations, old_block, manager)
        self.add_key_to_manager((b"key4",), locations, old_block, manager)
        action, _last_byte, _total_bytes = manager._check_rebuild_action()
        self.assertEqual("rebuild", action)
        manager._rebuild_block()
        new_block = manager._block
        self.assertIsNot(old_block, new_block)
        # Because of the new max_bytes_to_index, we do a poor job of
        # rebuilding. This is a side-effect of the change, but at least it does
        # show the setting had an effect.
        self.assertLess(old_block._content_length, new_block._content_length)

    def test_check_is_well_utilized_all_keys(self):
        block, manager = self.make_block_and_full_manager(self._texts)
        self.assertFalse(manager.check_is_well_utilized())
        # Though we can fake it by changing the recommended minimum size
        manager._full_enough_block_size = block._content_length
        self.assertTrue(manager.check_is_well_utilized())
        # Setting it just above causes it to fail
        manager._full_enough_block_size = block._content_length + 1
        self.assertFalse(manager.check_is_well_utilized())
        # Setting the mixed-block size doesn't do anything, because the content
        # is considered to not be 'mixed'
        manager._full_enough_mixed_block_size = block._content_length
        self.assertFalse(manager.check_is_well_utilized())

    def test_check_is_well_utilized_mixed_keys(self):
        texts = {}
        f1k1 = (b"f1", b"k1")
        f1k2 = (b"f1", b"k2")
        f2k1 = (b"f2", b"k1")
        f2k2 = (b"f2", b"k2")
        texts[f1k1] = self._texts[(b"key1",)]
        texts[f1k2] = self._texts[(b"key2",)]
        texts[f2k1] = self._texts[(b"key3",)]
        texts[f2k2] = self._texts[(b"key4",)]
        block, manager = self.make_block_and_full_manager(texts)
        self.assertFalse(manager.check_is_well_utilized())
        manager._full_enough_block_size = block._content_length
        self.assertTrue(manager.check_is_well_utilized())
        manager._full_enough_block_size = block._content_length + 1
        self.assertFalse(manager.check_is_well_utilized())
        manager._full_enough_mixed_block_size = block._content_length
        self.assertTrue(manager.check_is_well_utilized())

    def test_check_is_well_utilized_partial_use(self):
        locations, block = self.make_block(self._texts)
        manager = groupcompress._LazyGroupContentManager(block)
        manager._full_enough_block_size = block._content_length
        self.add_key_to_manager((b"key1",), locations, block, manager)
        self.add_key_to_manager((b"key2",), locations, block, manager)
        # Just using the content from key1 and 2 is not enough to be considered
        # 'complete'
        self.assertFalse(manager.check_is_well_utilized())
        # However if we add key3, then we have enough, as we only require 75%
        # consumption
        self.add_key_to_manager((b"key4",), locations, block, manager)
        self.assertTrue(manager.check_is_well_utilized())


class Test_GCBuildDetails(TestCase):
    def test_acts_like_tuple(self):
        # _GCBuildDetails inlines some of the data that used to be spread out
        # across a bunch of tuples
        bd = groupcompress._GCBuildDetails(
            (("parent1",), ("parent2",)), ("INDEX", 10, 20, 0, 5)
        )
        self.assertEqual(4, len(bd))
        self.assertEqual(("INDEX", 10, 20, 0, 5), bd[0])
        self.assertEqual(None, bd[1])  # Compression Parent is always None
        self.assertEqual((("parent1",), ("parent2",)), bd[2])
        self.assertEqual(("group", None), bd[3])  # Record details

    def test__repr__(self):
        bd = groupcompress._GCBuildDetails(
            (("parent1",), ("parent2",)), ("INDEX", 10, 20, 0, 5)
        )
        self.assertEqual(
            "_GCBuildDetails(('INDEX', 10, 20, 0, 5), (('parent1',), ('parent2',)))",
            repr(bd),
        )


class TestBase128Int(TestCase):
    def assertEqualEncode(self, bytes, val):
        self.assertEqual(bytes, groupcompress.encode_base128_int(val))

    def assertEqualDecode(self, val, num_decode, bytes):
        self.assertEqual((val, num_decode), groupcompress.decode_base128_int(bytes))

    def test_encode(self):
        self.assertEqualEncode(b"\x01", 1)
        self.assertEqualEncode(b"\x02", 2)
        self.assertEqualEncode(b"\x7f", 127)
        self.assertEqualEncode(b"\x80\x01", 128)
        self.assertEqualEncode(b"\xff\x01", 255)
        self.assertEqualEncode(b"\x80\x02", 256)
        self.assertEqualEncode(b"\xff\xff\xff\xff\x0f", 0xFFFFFFFF)

    def test_decode(self):
        self.assertEqualDecode(1, 1, b"\x01")
        self.assertEqualDecode(2, 1, b"\x02")
        self.assertEqualDecode(127, 1, b"\x7f")
        self.assertEqualDecode(128, 2, b"\x80\x01")
        self.assertEqualDecode(255, 2, b"\xff\x01")
        self.assertEqualDecode(256, 2, b"\x80\x02")
        self.assertEqualDecode(0xFFFFFFFF, 5, b"\xff\xff\xff\xff\x0f")

    def test_decode_with_trailing_bytes(self):
        self.assertEqualDecode(1, 1, b"\x01abcdef")
        self.assertEqualDecode(127, 1, b"\x7f\x01")
        self.assertEqualDecode(128, 2, b"\x80\x01abcdef")
        self.assertEqualDecode(255, 2, b"\xff\x01\xff")



# Copyright (C) 2025 Breezy Contributors
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""File locking for bzrformats.

Uses fcntl.lockf which has per-process semantics: multiple file descriptors
within the same process can share a lock on the same file.
"""

import fcntl

from .errors import LockContention


class ReadLock:
    """OS-level shared (read) lock on a file.

    The locked file is accessible via the ``f`` attribute.
    """

    def __init__(self, filename):
        """Acquire a shared read lock on *filename*."""
        self.filename = filename
        self.f = open(filename, "rb")
        try:
            fcntl.lockf(self.f, fcntl.LOCK_SH | fcntl.LOCK_NB)
        except BlockingIOError:
            self.f.close()
            raise LockContention(filename) from None
        except OSError:
            self.f.close()
            raise

    def unlock(self):
        """Release the lock and close the file."""
        fcntl.lockf(self.f, fcntl.LOCK_UN)
        self.f.close()

    def temporary_write_lock(self):
        """Try to upgrade to a write lock.

        Returns ``(True, write_lock)`` if the upgrade succeeded, or
        ``(False, self)`` if it failed (read lock is re-acquired).
        """
        fcntl.lockf(self.f, fcntl.LOCK_UN)
        self.f.close()
        try:
            wl = WriteLock(self.filename)
            return True, wl
        except (LockContention, OSError):
            # Re-acquire read lock
            self.f = open(self.filename, "rb")
            fcntl.lockf(self.f, fcntl.LOCK_SH | fcntl.LOCK_NB)
            return False, self


class WriteLock:
    """OS-level exclusive (write) lock on a file.

    The locked file is accessible via the ``f`` attribute.
    Creates the file if it does not exist.
    """

    def __init__(self, filename):
        """Acquire an exclusive write lock on *filename*."""
        self.filename = filename
        try:
            self.f = open(filename, "rb+")
        except FileNotFoundError:
            self.f = open(filename, "wb+")
        try:
            fcntl.lockf(self.f, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except BlockingIOError:
            self.f.close()
            raise LockContention(filename) from None
        except OSError:
            self.f.close()
            raise

    def unlock(self):
        """Release the lock and close the file."""
        fcntl.lockf(self.f, fcntl.LOCK_UN)
        self.f.close()

    def restore_read_lock(self):
        """Downgrade to a read lock, returning a new :class:`ReadLock`."""
        fcntl.lockf(self.f, fcntl.LOCK_UN)
        self.f.close()
        return ReadLock(self.filename)


class LogicalLockResult:
    """The result of a lock_read/lock_write call.

    Can be used as a context manager::

        with tree.lock_read():
            ...
    """

    def __init__(self, unlock, token=None):
        """Initialize with an unlock callable and optional token."""
        self.unlock = unlock
        self.token = token

    def __repr__(self):
        """Return string representation."""
        return f"LogicalLockResult({self.unlock})"

    def __enter__(self):
        """Enter context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit context manager, releasing the lock."""
        try:
            self.unlock()
        except BaseException:
            if exc_type is None:
                raise
        return False



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the compiled dirstate helpers."""

import bisect
import os

from testscenarios import load_tests_apply_scenarios

from .. import _dirstate_helpers_py, dirstate
from . import TestCase

load_tests = load_tests_apply_scenarios

try:
    from .. import _dirstate_helpers_pyx as compiled_dirstate_helpers
except ImportError:
    compiled_dirstate_helpers = None


helper_scenarios = [("dirstate_Python", {"helpers": _dirstate_helpers_py})]
if compiled_dirstate_helpers is not None:
    helper_scenarios.append(("dirstate_Pyrex", {"helpers": compiled_dirstate_helpers}))


class TestBisectPathMixin:
    """Test that _bisect_path_*() returns the expected values.

    _bisect_path_* is intended to work like bisect.bisect_*() except it
    knows it is working on paths that are sorted by ('path', 'to', 'foo')
    chunks rather than by raw 'path/to/foo'.

    Test Cases should inherit from this and override ``get_bisect_path`` return
    their implementation, and ``get_bisect`` to return the matching
    bisect.bisect_* function.
    """

    def get_bisect_path(self):
        """Return an implementation of _bisect_path_*."""
        raise NotImplementedError

    def get_bisect(self):
        """Return a version of bisect.bisect_*.

        Also, for the 'exists' check, return the offset to the real values.
        For example bisect_left returns the index of an entry, while
        bisect_right returns the index *after* an entry

        :return: (bisect_func, offset)
        """
        raise NotImplementedError

    def assertBisect(self, paths, split_paths, path, exists=True):
        """Assert that bisect_split works like bisect_left on the split paths.

        :param paths: A list of path names
        :param split_paths: A list of path names that are already split up by directory
            ('path/to/foo' => ('path', 'to', 'foo'))
        :param path: The path we are indexing.
        :param exists: The path should be present, so make sure the
            final location actually points to the right value.

        All other arguments will be passed along.
        """
        bisect_path = self.get_bisect_path()
        self.assertIsInstance(paths, list)
        bisect_path_idx = bisect_path(paths, path)
        split_path = self.split_for_dirblocks([path])[0]
        bisect_func, offset = self.get_bisect()
        bisect_split_idx = bisect_func(split_paths, split_path)
        self.assertEqual(
            bisect_split_idx,
            bisect_path_idx,
            "{} disagreed. {} != {} for key {!r}".format(
                bisect_path.__name__, bisect_split_idx, bisect_path_idx, path
            ),
        )
        if exists:
            self.assertEqual(path, paths[bisect_path_idx + offset])

    def split_for_dirblocks(self, paths):
        dir_split_paths = []
        for path in paths:
            dirname, basename = os.path.split(path)
            dir_split_paths.append((dirname.split(b"/"), basename))
        dir_split_paths.sort()
        return dir_split_paths

    def test_simple(self):
        """In the simple case it works just like bisect_left."""
        paths = [b"", b"a", b"b", b"c", b"d"]
        split_paths = self.split_for_dirblocks(paths)
        for path in paths:
            self.assertBisect(paths, split_paths, path, exists=True)
        self.assertBisect(paths, split_paths, b"_", exists=False)
        self.assertBisect(paths, split_paths, b"aa", exists=False)
        self.assertBisect(paths, split_paths, b"bb", exists=False)
        self.assertBisect(paths, split_paths, b"cc", exists=False)
        self.assertBisect(paths, split_paths, b"dd", exists=False)
        self.assertBisect(paths, split_paths, b"a/a", exists=False)
        self.assertBisect(paths, split_paths, b"b/b", exists=False)
        self.assertBisect(paths, split_paths, b"c/c", exists=False)
        self.assertBisect(paths, split_paths, b"d/d", exists=False)

    def test_involved(self):
        """This is where bisect_path_* diverges slightly."""
        # This is the list of paths and their contents
        # a/
        #   a/
        #     a
        #     z
        #   a-a/
        #     a
        #   a-z/
        #     z
        #   a=a/
        #     a
        #   a=z/
        #     z
        #   z/
        #     a
        #     z
        #   z-a
        #   z-z
        #   z=a
        #   z=z
        # a-a/
        #   a
        # a-z/
        #   z
        # a=a/
        #   a
        # a=z/
        #   z
        # This is the exact order that is stored by dirstate
        # All children in a directory are mentioned before an children of
        # children are mentioned.
        # So all the root-directory paths, then all the
        # first sub directory, etc.
        paths = [  # content of '/'
            b"",
            b"a",
            b"a-a",
            b"a-z",
            b"a=a",
            b"a=z",
            # content of 'a/'
            b"a/a",
            b"a/a-a",
            b"a/a-z",
            b"a/a=a",
            b"a/a=z",
            b"a/z",
            b"a/z-a",
            b"a/z-z",
            b"a/z=a",
            b"a/z=z",
            # content of 'a/a/'
            b"a/a/a",
            b"a/a/z",
            # content of 'a/a-a'
            b"a/a-a/a",
            # content of 'a/a-z'
            b"a/a-z/z",
            # content of 'a/a=a'
            b"a/a=a/a",
            # content of 'a/a=z'
            b"a/a=z/z",
            # content of 'a/z/'
            b"a/z/a",
            b"a/z/z",
            # content of 'a-a'
            b"a-a/a",
            # content of 'a-z'
            b"a-z/z",
            # content of 'a=a'
            b"a=a/a",
            # content of 'a=z'
            b"a=z/z",
        ]
        split_paths = self.split_for_dirblocks(paths)
        sorted_paths = []
        for dir_parts, basename in split_paths:
            if dir_parts == [b""]:
                sorted_paths.append(basename)
            else:
                sorted_paths.append(b"/".join(dir_parts + [basename]))

        self.assertEqual(sorted_paths, paths)

        for path in paths:
            self.assertBisect(paths, split_paths, path, exists=True)


class TestBisectPathLeft(TestCase, TestBisectPathMixin):
    """Run all Bisect Path tests against bisect_path_left."""

    def get_bisect_path(self):
        from ..dirstate import bisect_path_left

        return bisect_path_left

    def get_bisect(self):
        return bisect.bisect_left, 0


class TestBisectPathRight(TestCase, TestBisectPathMixin):
    """Run all Bisect Path tests against bisect_path_right."""

    def get_bisect_path(self):
        from ..dirstate import bisect_path_right

        return bisect_path_right

    def get_bisect(self):
        return bisect.bisect_right, -1


class TestLtByDirs(TestCase):
    """Test an implementation of lt_by_dirs().

    lt_by_dirs() compares 2 paths by their directory sections, rather than as
    plain strings.
    """

    def assertCmpByDirs(self, expected, str1, str2):
        """Compare the two strings, in both directions.

        :param expected: The expected comparison value. -1 means str1 comes
            first, 0 means they are equal, 1 means str2 comes first
        :param str1: string to compare
        :param str2: string to compare
        """
        if expected == 0:
            self.assertEqual(str1, str2)
            self.assertFalse(dirstate.lt_by_dirs(str1, str2))
            self.assertFalse(dirstate.lt_by_dirs(str2, str1))
        elif expected > 0:
            self.assertFalse(dirstate.lt_by_dirs(str1, str2))
            self.assertTrue(dirstate.lt_by_dirs(str2, str1))
        else:
            self.assertTrue(dirstate.lt_by_dirs(str1, str2))
            self.assertFalse(dirstate.lt_by_dirs(str2, str1))

    def test_cmp_empty(self):
        """Compare against the empty string."""
        self.assertCmpByDirs(0, b"", b"")
        self.assertCmpByDirs(1, b"a", b"")
        self.assertCmpByDirs(1, b"ab", b"")
        self.assertCmpByDirs(1, b"abc", b"")
        self.assertCmpByDirs(1, b"abcd", b"")
        self.assertCmpByDirs(1, b"abcde", b"")
        self.assertCmpByDirs(1, b"abcdef", b"")
        self.assertCmpByDirs(1, b"abcdefg", b"")
        self.assertCmpByDirs(1, b"abcdefgh", b"")
        self.assertCmpByDirs(1, b"abcdefghi", b"")
        self.assertCmpByDirs(1, b"test/ing/a/path/", b"")

    def test_cmp_same_str(self):
        """Compare the same string."""
        self.assertCmpByDirs(0, b"a", b"a")
        self.assertCmpByDirs(0, b"ab", b"ab")
        self.assertCmpByDirs(0, b"abc", b"abc")
        self.assertCmpByDirs(0, b"abcd", b"abcd")
        self.assertCmpByDirs(0, b"abcde", b"abcde")
        self.assertCmpByDirs(0, b"abcdef", b"abcdef")
        self.assertCmpByDirs(0, b"abcdefg", b"abcdefg")
        self.assertCmpByDirs(0, b"abcdefgh", b"abcdefgh")
        self.assertCmpByDirs(0, b"abcdefghi", b"abcdefghi")
        self.assertCmpByDirs(0, b"testing a long string", b"testing a long string")
        self.assertCmpByDirs(0, b"x" * 10000, b"x" * 10000)
        self.assertCmpByDirs(0, b"a/b", b"a/b")
        self.assertCmpByDirs(0, b"a/b/c", b"a/b/c")
        self.assertCmpByDirs(0, b"a/b/c/d", b"a/b/c/d")
        self.assertCmpByDirs(0, b"a/b/c/d/e", b"a/b/c/d/e")

    def test_simple_paths(self):
        """Compare strings that act like normal string comparison."""
        self.assertCmpByDirs(-1, b"a", b"b")
        self.assertCmpByDirs(-1, b"aa", b"ab")
        self.assertCmpByDirs(-1, b"ab", b"bb")
        self.assertCmpByDirs(-1, b"aaa", b"aab")
        self.assertCmpByDirs(-1, b"aab", b"abb")
        self.assertCmpByDirs(-1, b"abb", b"bbb")
        self.assertCmpByDirs(-1, b"aaaa", b"aaab")
        self.assertCmpByDirs(-1, b"aaab", b"aabb")
        self.assertCmpByDirs(-1, b"aabb", b"abbb")
        self.assertCmpByDirs(-1, b"abbb", b"bbbb")
        self.assertCmpByDirs(-1, b"aaaaa", b"aaaab")
        self.assertCmpByDirs(-1, b"a/a", b"a/b")
        self.assertCmpByDirs(-1, b"a/b", b"b/b")
        self.assertCmpByDirs(-1, b"a/a/a", b"a/a/b")
        self.assertCmpByDirs(-1, b"a/a/b", b"a/b/b")
        self.assertCmpByDirs(-1, b"a/b/b", b"b/b/b")
        self.assertCmpByDirs(-1, b"a/a/a/a", b"a/a/a/b")
        self.assertCmpByDirs(-1, b"a/a/a/b", b"a/a/b/b")
        self.assertCmpByDirs(-1, b"a/a/b/b", b"a/b/b/b")
        self.assertCmpByDirs(-1, b"a/b/b/b", b"b/b/b/b")
        self.assertCmpByDirs(-1, b"a/a/a/a/a", b"a/a/a/a/b")

    def test_tricky_paths(self):
        self.assertCmpByDirs(1, b"ab/cd/ef", b"ab/cc/ef")
        self.assertCmpByDirs(1, b"ab/cd/ef", b"ab/c/ef")
        self.assertCmpByDirs(-1, b"ab/cd/ef", b"ab/cd-ef")
        self.assertCmpByDirs(-1, b"ab/cd", b"ab/cd-")
        self.assertCmpByDirs(-1, b"ab/cd", b"ab-cd")

    def test_cmp_non_ascii(self):
        self.assertCmpByDirs(-1, b"\xc2\xb5", b"\xc3\xa5")  # u'\xb5', u'\xe5'
        self.assertCmpByDirs(-1, b"a", b"\xc3\xa5")  # u'a', u'\xe5'
        self.assertCmpByDirs(-1, b"b", b"\xc2\xb5")  # u'b', u'\xb5'
        self.assertCmpByDirs(-1, b"a/b", b"a/\xc3\xa5")  # u'a/b', u'a/\xe5'
        self.assertCmpByDirs(-1, b"b/a", b"b/\xc2\xb5")  # u'b/a', u'b/\xb5'


class TestLtPathByDirblock(TestCase):
    """Test an implementation of lt_path_by_dirblock().

    lt_path_by_dirblock() compares two paths using the sort order used by
    DirState. All paths in the same directory are sorted together.

    Child test cases can override ``get_lt_path_by_dirblock`` to test a specific
    implementation.
    """

    def get_lt_path_by_dirblock(self):
        """Get a specific implementation of lt_path_by_dirblock."""
        from ..dirstate import lt_path_by_dirblock

        return lt_path_by_dirblock

    def assertLtPathByDirblock(self, paths):
        """Compare all paths and make sure they evaluate to the correct order.

        This does N^2 comparisons. It is assumed that ``paths`` is properly
        sorted list.

        :param paths: a sorted list of paths to compare
        """

        # First, make sure the paths being passed in are correct
        def _key(p):
            dirname, basename = os.path.split(p)
            return dirname.split(b"/"), basename

        self.assertEqual(sorted(paths, key=_key), paths)

        lt_path_by_dirblock = self.get_lt_path_by_dirblock()
        for idx1, path1 in enumerate(paths):
            for idx2, path2 in enumerate(paths):
                lt_result = lt_path_by_dirblock(path1, path2)
                self.assertEqual(
                    idx1 < idx2,
                    lt_result,
                    "{} did not state that {!r} < {!r}, lt={}".format(
                        lt_path_by_dirblock.__name__, path1, path2, lt_result
                    ),
                )

    def test_cmp_simple_paths(self):
        """Compare against the empty string."""
        self.assertLtPathByDirblock([b"", b"a", b"ab", b"abc", b"a/b/c", b"b/d/e"])
        self.assertLtPathByDirblock([b"kl", b"ab/cd", b"ab/ef", b"gh/ij"])

    def test_tricky_paths(self):
        self.assertLtPathByDirblock(
            [
                # Contents of ''
                b"",
                b"a",
                b"a-a",
                b"a=a",
                b"b",
                # Contents of 'a'
                b"a/a",
                b"a/a-a",
                b"a/a=a",
                b"a/b",
                # Contents of 'a/a'
                b"a/a/a",
                b"a/a/a-a",
                b"a/a/a=a",
                # Contents of 'a/a/a'
                b"a/a/a/a",
                b"a/a/a/b",
                # Contents of 'a/a/a-a',
                b"a/a/a-a/a",
                b"a/a/a-a/b",
                # Contents of 'a/a/a=a',
                b"a/a/a=a/a",
                b"a/a/a=a/b",
                # Contents of 'a/a-a'
                b"a/a-a/a",
                # Contents of 'a/a-a/a'
                b"a/a-a/a/a",
                b"a/a-a/a/b",
                # Contents of 'a/a=a'
                b"a/a=a/a",
                # Contents of 'a/b'
                b"a/b/a",
                b"a/b/b",
                # Contents of 'a-a',
                b"a-a/a",
                b"a-a/b",
                # Contents of 'a=a',
                b"a=a/a",
                b"a=a/b",
                # Contents of 'b',
                b"b/a",
                b"b/b",
            ]
        )
        self.assertLtPathByDirblock(
            [
                # content of '/'
                b"",
                b"a",
                b"a-a",
                b"a-z",
                b"a=a",
                b"a=z",
                # content of 'a/'
                b"a/a",
                b"a/a-a",
                b"a/a-z",
                b"a/a=a",
                b"a/a=z",
                b"a/z",
                b"a/z-a",
                b"a/z-z",
                b"a/z=a",
                b"a/z=z",
                # content of 'a/a/'
                b"a/a/a",
                b"a/a/z",
                # content of 'a/a-a'
                b"a/a-a/a",
                # content of 'a/a-z'
                b"a/a-z/z",
                # content of 'a/a=a'
                b"a/a=a/a",
                # content of 'a/a=z'
                b"a/a=z/z",
                # content of 'a/z/'
                b"a/z/a",
                b"a/z/z",
                # content of 'a-a'
                b"a-a/a",
                # content of 'a-z'
                b"a-z/z",
                # content of 'a=a'
                b"a=a/a",
                # content of 'a=z'
                b"a=z/z",
            ]
        )

    def test_nonascii(self):
        self.assertLtPathByDirblock(
            [
                # content of '/'
                b"",
                b"a",
                b"\xc2\xb5",
                b"\xc3\xa5",
                # content of 'a'
                b"a/a",
                b"a/\xc2\xb5",
                b"a/\xc3\xa5",
                # content of 'a/a'
                b"a/a/a",
                b"a/a/\xc2\xb5",
                b"a/a/\xc3\xa5",
                # content of 'a/\xc2\xb5'
                b"a/\xc2\xb5/a",
                b"a/\xc2\xb5/\xc2\xb5",
                b"a/\xc2\xb5/\xc3\xa5",
                # content of 'a/\xc3\xa5'
                b"a/\xc3\xa5/a",
                b"a/\xc3\xa5/\xc2\xb5",
                b"a/\xc3\xa5/\xc3\xa5",
                # content of '\xc2\xb5'
                b"\xc2\xb5/a",
                b"\xc2\xb5/\xc2\xb5",
                b"\xc2\xb5/\xc3\xa5",
                # content of '\xc2\xe5'
                b"\xc3\xa5/a",
                b"\xc3\xa5/\xc2\xb5",
                b"\xc3\xa5/\xc3\xa5",
            ]
        )


class TestUsingCompiledIfAvailable(TestCase):
    """Check that any compiled functions that are available are the default.

    It is possible to have typos, etc in the import line, such that
    _dirstate_helpers_pyx is actually available, but the compiled functions are
    not being used.
    """

    def test__read_dirblocks(self):
        if compiled_dirstate_helpers is not None:
            from .._dirstate_helpers_pyx import _read_dirblocks
        else:
            from .._dirstate_helpers_py import _read_dirblocks
        self.assertIs(_read_dirblocks, dirstate._read_dirblocks)

    def test_update_entry(self):
        if compiled_dirstate_helpers is not None:
            from .._dirstate_helpers_pyx import update_entry
        else:
            from ..dirstate import update_entry
        self.assertIs(update_entry, dirstate.update_entry)

    def test_process_entry(self):
        if compiled_dirstate_helpers is not None:
            from .._dirstate_helpers_pyx import ProcessEntryC

            self.assertIs(ProcessEntryC, dirstate._process_entry)
        else:
            from ..dirstate import ProcessEntryPython

            self.assertIs(ProcessEntryPython, dirstate._process_entry)



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Serializer object for CHK based inventory storage."""

from . import serializer


class CHKSerializer(serializer.InventorySerializer):
    """A CHKInventory based serializer with 'plain' behaviour."""

    support_altered_by_hack = False
    supported_kinds = {"file", "directory", "symlink", "tree-reference"}

    def __init__(self, format_num, node_size, search_key_name):
        """Initialize a CHKSerializer instance.

        Args:
            format_num: The format number for the serializer (e.g., b"9" or b"10").
            node_size: The maximum size for CHK nodes (typically 65536).
            search_key_name: The name of the search key algorithm (e.g., b"hash-255-way").
        """
        self.format_num = format_num
        self.maximum_size = node_size
        self.search_key_name = search_key_name

    def _unpack_inventory(
        self, elt, revision_id=None, entry_cache=None, return_from_cache=False
    ):
        """Construct from XML Element."""
        from .xml_serializer import unpack_inventory_entry, unpack_inventory_flat

        inv = unpack_inventory_flat(
            elt, self.format_num, unpack_inventory_entry, entry_cache, return_from_cache
        )
        return inv

    def read_inventory_from_lines(
        self, xml_lines, revision_id=None, entry_cache=None, return_from_cache=False
    ):
        """Read xml_string into an inventory object.

        :param xml_string: The xml to read.
        :param revision_id: If not-None, the expected revision id of the
            inventory.
        :param entry_cache: An optional cache of InventoryEntry objects. If
            supplied we will look up entries via (file_id, revision_id) which
            should map to a valid InventoryEntry (File/Directory/etc) object.
        :param return_from_cache: Return entries directly from the cache,
            rather than copying them first. This is only safe if the caller
            promises not to mutate the returned inventory entries, but it can
            make some operations significantly faster.
        """
        from .xml_serializer import ParseError, fromstringlist

        try:
            return self._unpack_inventory(
                fromstringlist(xml_lines),
                revision_id,
                entry_cache=entry_cache,
                return_from_cache=return_from_cache,
            )
        except ParseError as e:
            raise serializer.UnexpectedInventoryFormat(e) from e

    def read_inventory(self, f, revision_id=None):
        """Read an inventory from a file-like object."""
        from .xml_serializer import ParseError

        try:
            try:
                return self._unpack_inventory(self._read_element(f), revision_id=None)
            finally:
                f.close()
        except ParseError as e:
            raise serializer.UnexpectedInventoryFormat(e) from e

    def write_inventory_to_lines(self, inv):
        """Return a list of lines with the encoded inventory."""
        return self.write_inventory(inv, None)

    def write_inventory_to_chunks(self, inv):
        """Return a list of lines with the encoded inventory."""
        return self.write_inventory(inv, None)

    def write_inventory(self, inv, f, working=False):
        """Write inventory to a file.

        :param inv: the inventory to write.
        :param f: the file to write. (May be None if the lines are the desired
            output).
        :param working: If True skip history data - text_sha1, text_size,
            reference_revision, symlink_target.
        :return: The inventory as a list of lines.
        """
        from .xml_serializer import encode_and_escape, serialize_inventory_flat

        output = []
        append = output.append
        if inv.revision_id is not None:
            revid = b"".join(
                [b' revision_id="', encode_and_escape(inv.revision_id), b'"']
            )
        else:
            revid = b""
        append(b'<inventory format="%s"%s>\n' % (self.format_num, revid))
        append(
            b'<directory file_id="%s" name="%s" revision="%s" />\n'
            % (
                encode_and_escape(inv.root.file_id),
                encode_and_escape(inv.root.name),
                encode_and_escape(inv.root.revision),
            )
        )
        serialize_inventory_flat(
            inv,
            append,
            root_id=None,
            supported_kinds=self.supported_kinds,
            working=working,
        )
        if f is not None:
            f.writelines(output)
        return output


# A CHKInventory based serializer with 'plain' behaviour.
inventory_chk_serializer_255_bigpage_9 = CHKSerializer(b"9", 65536, b"hash-255-way")
inventory_chk_serializer_255_bigpage_10 = CHKSerializer(b"10", 65536, b"hash-255-way")



# Copyright (C) 2004-2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Diff utilities for bzrformats."""

import difflib


class _PrematchedMatcher(difflib.SequenceMatcher):
    """Allow SequenceMatcher operations to use predetermined blocks."""

    def __init__(self, matching_blocks):
        difflib.SequenceMatcher(self, None, None)
        self.matching_blocks = matching_blocks
        self.opcodes = None
